import os
import requests
import argparse
from tqdm import tqdm

def download_tile(z, x, y, url_template, output_folder, pbar):
    """Download a single tile and save it in a structured folder (z/x/y.pbf)."""
    tile_url = url_template.format(z=z, x=x, y=y)
    response = requests.get(tile_url)
    if response.status_code == 200:
        # Define the path structure: output_folder/z/x/y.pbf
        tile_folder = os.path.join(output_folder, str(z), str(x))
        os.makedirs(tile_folder, exist_ok=True)  # Create directories if they don't exist
        tile_file = os.path.join(tile_folder, f'{y}.pbf')  # Save file as y.pbf
        with open(tile_file, 'wb') as f:
            f.write(response.content)
    pbar.update(1)  # Update the progress bar after each download

def download_tiles(minzoom, maxzoom, url_template, output_folder):
    """Download all tiles from minzoom to maxzoom in a structured folder with a progress bar."""
    total_tiles = sum((2 ** z) ** 2 for z in range(minzoom, maxzoom + 1))
    with tqdm(total=total_tiles, desc="Downloading tiles", unit="tile") as pbar:
        for z in range(minzoom, maxzoom + 1):
            num_tiles = 2 ** z
            for x in range(num_tiles):
                for y in range(num_tiles):
                    download_tile(z, x, y, url_template, output_folder, pbar)

def main():
    parser = argparse.ArgumentParser(description='Download vector tiles from a tile server and store them in structured folders.')
    parser.add_argument('-url', type=str, required=True, help='URL template for vector tiles (e.g., https://your-vector-tile-server/{z}/{x}/{y}.pbf)')
    parser.add_argument('-minzoom', type=int, required=True, help='Minimum zoom level')
    parser.add_argument('-maxzoom', type=int, required=True, help='Maximum zoom level')
    parser.add_argument('-o', type=str, default='tiles', help='Output folder for downloaded tiles')

    args = parser.parse_args()

    download_tiles(args.minzoom, args.maxzoom, args.url, args.o)

if __name__ == '__main__':
    main()



import json, math
import argparse
from vtiles.utils.mapbox_vector_tile import encode
import vtiles.utils.mercantile as mercantile
import sqlite3
from shapely.geometry import shape, mapping
from shapely.ops import transform
from shapely.wkt import dumps as wkt_dumps
import gzip
from vtiles.mbtiles import mbtilesfixmeta
import pyproj

def create_mbtiles(db_path):
    conn = sqlite3.connect(db_path)
    cursor = conn.cursor()
    cursor.execute('CREATE TABLE metadata (name TEXT, value TEXT);')
    cursor.execute('CREATE UNIQUE INDEX name on metadata (name);')
    cursor.execute('CREATE TABLE tiles (zoom_level INTEGER, tile_column INTEGER, tile_row INTEGER, tile_data BLOB);')
    cursor.execute('CREATE UNIQUE INDEX tile_index ON tiles (zoom_level, tile_column, tile_row);')
    conn.commit()
    conn.close()

def geojson_to_custom_format(geojson):
    layer_name = geojson.get('name', 'geojson_layer')
    layers = {}
    # Process each feature
    for feature in geojson['features']:
        if layer_name not in layers:
            layers[layer_name] = []

        # Convert GeoJSON feature geometry to WKT format
        geom = shape(feature['geometry'])  # Convert GeoJSON to Shapely geometry
        geom_wkt = wkt_dumps(geom)  # Convert Shapely geometry to WKT

        # Prepare feature dictionary
        feature_dict = {
            "geometry": geom_wkt,
            "properties": feature["properties"]
        }

        # Add feature to the appropriate layer
        layers[layer_name].append(feature_dict)

    # Convert layers dictionary to the desired format
    formatted_layers = [
        {
            "name": layer_name,
            "features": features
        }
        for layer_name, features in layers.items()
    ]
    return formatted_layers

# def reproject_geojson_to_tile_extent(geojson_data, tile_x, tile_y, zoom, tile_size=4096):
#     # Get the bounds of the specified tile
#     tile_bounds = mercantile.bounds(tile_x, tile_y, zoom)

#     # Calculate scale factors
#     minx, miny, maxx, maxy = tile_bounds
#     scale_x = tile_size / (maxx - minx)
#     scale_y = tile_size / (maxy - miny)

#     def scale_coords(x, y, z=None):
#         return (x - minx) * scale_x, (y - miny) * scale_y

#     # Convert and scale features
#     reprojected_features = []
#     for feature in geojson_data["features"]:
#         geometry = shape(feature["geometry"])
#         scaled_geometry = transform(scale_coords, geometry)
#         reprojected_features.append({
#             "type": "Feature",
#             "properties": feature["properties"],
#             "geometry": mapping(scaled_geometry)
#         })

#     # Create the new GeoJSON structure
#     reprojected_geojson = {
#         "type": "FeatureCollection",
#         "name": geojson_data.get("name", "reprojected_geojson"),
#         "crs": geojson_data.get("crs"),
#         "features": reprojected_features
#     }

#     return reprojected_geojson

def reproject_geojson_to_tile_extent(geojson_data, tile_x, tile_y, zoom, tile_size=4096):
    # Define the projections
    proj_wgs84 = pyproj.Transformer.from_crs("EPSG:4326", "EPSG:3857", always_xy=True)
    proj_web_mercator = pyproj.Transformer.from_crs("EPSG:3857", "EPSG:3857", always_xy=True)

    # Get the bounds of the specified tile in Web Mercator coordinates
    tile_bounds = mercantile.xy_bounds(tile_x, tile_y, zoom)
    # tile_bounds = mercantile.bounds(tile_x, tile_y, zoom)

    # Calculate scale factors
    minx, miny, maxx, maxy = tile_bounds
    if maxx == minx or maxy == miny:
        raise ValueError("Invalid tile bounds: zero-width or zero-height tile")
    scale_x = tile_size / (maxx - minx)
    scale_y = tile_size / (maxy - miny)

    # Define the transformation functions
    def reproject_geom(geometry):
        return transform(lambda x, y: proj_wgs84.transform(x, y), geometry)

    def scale_coords(x, y, z=None):
        x_scaled = (x - minx) * scale_x
        # y_scaled = tile_size - (y - miny) * scale_y  # Flip y-axis
        y_scaled = (y - miny) * scale_y   
        return x_scaled, y_scaled
        

    def scale_geom(geometry):
        # return transform(lambda x, y: scale_coords(x, y), geometry)
        return transform(lambda x, y: scale_coords(x, y), geometry)

    # Reproject and scale features
    reprojected_features = []
    for feature in geojson_data["features"]:
        geometry = shape(feature["geometry"])
        
        # Reproject from WGS84 to Web Mercator
        reprojected_geometry = reproject_geom(geometry)
        
        # Scale the reprojected geometry to tile extent
        scaled_geometry = scale_geom(reprojected_geometry)
        
        # Store the reprojected geometry
        reprojected_features.append({
            "type": "Feature",
            "properties": feature["properties"],
            "geometry": mapping(scaled_geometry)
        })

    # Create the new GeoJSON structure
    reprojected_geojson = {
        "type": "FeatureCollection",
        "name": geojson_data.get("name", "reprojected_geojson"),
        "crs": geojson_data.get("crs"),
        "features": reprojected_features
    }

    return reprojected_geojson

def add_tile_to_mbtiles(db_path, z, x, y, tile_data):
    """Add a tile to the MBTiles database."""
    conn = sqlite3.connect(db_path)
    c = conn.cursor()
    c.execute('INSERT OR REPLACE INTO tiles (zoom_level, tile_column, tile_row, tile_data) VALUES (?, ?, ?, ?);', (z, x, y, tile_data))
    conn.commit()
    conn.close()

def main():
    parser = argparse.ArgumentParser(description="Convert GeoJSON to MBTiles.")
    parser.add_argument('-i', '--input', required=True, help="Input GeoJSON file.")
    parser.add_argument('-o', '--output', required=True, help="Output MBTiles file.")
    parser.add_argument('-z', '--zoom', type=int, default=0, help="Zoom level for the tile.")
    parser.add_argument('-x', '--x', type=int, default=0, help="Tile column.")
    parser.add_argument('-y', '--y', type=int, default=0, help="Tile row.")
    args = parser.parse_args()

    # Read the input GeoJSON file
    with open(args.input, 'r',encoding='utf-8') as f:
        geojson_data = json.load(f)

    # Define tile coordinates
    z, x, y = args.zoom, args.x, args.y

    # Create MBTiles file
    create_mbtiles(args.output)
    
    geojson_reprojected = reproject_geojson_to_tile_extent(geojson_data,z, x, y)
    # print (geojson_reprojected)
    geojson_reprojected_processed = geojson_to_custom_format(geojson_reprojected)
    print (geojson_reprojected_processed[0]['name'])
    geojson_reprojected_processed_encoded = encode(geojson_reprojected_processed)
    geojson_reprojected_processed_encoded_compressed = gzip.compress(geojson_reprojected_processed_encoded)

    # Add tile to MBTiles
    add_tile_to_mbtiles(args.output, z, x, y, geojson_reprojected_processed_encoded_compressed)
    mbtilesfixmeta.fix_metadata(args.output, 'GZIP') 

if __name__ == "__main__":
    main()



def create_feature(id_, type_, geom, tags):
    feature = {
        "id": None if id_ is None else str(id_),
        "type": type_,
        "geometry": geom,
        "tags": tags,
        "minX": float('inf'),
        "minY": float('inf'),
        "maxX": float('-inf'),
        "maxY": float('-inf')
    }

    if type_ == 'Point' or type_ == 'MultiPoint' or type_ == 'LineString':
        calc_line_bbox(feature, geom)
    elif type_ == 'Polygon':
        # the outer ring(ie[0]) contains all inner rings
        calc_line_bbox(feature, geom[0])
    elif type_ == 'MultiLineString':
        for line in geom:
            calc_line_bbox(feature, line)
    elif type_ == 'MultiPolygon':
        for polygon in geom:
            # the outer ring(ie[0]) contains all inner rings
            calc_line_bbox(feature, polygon[0])
    return feature


def calc_line_bbox(feature, geom):
    for i in range(0, len(geom), 3):
        feature['minX'] = min(feature.get('minX'), geom[i])
        feature['minY'] = min(feature.get('minY'), geom[i + 1])
        feature['maxX'] = max(feature.get('maxX'), geom[i])
        feature['maxY'] = max(feature.get('maxY'), geom[i + 1])


class Slice(list):
    def __init__(self, *args):
        list.__init__(self, *args)
        self.start = 0.
        self.end = 0.
        self.size = 0.



from itertools import combinations
import logging
import networkx as nx
from networkx.exception import NetworkXNoPath
import numpy as np
import operator
from scipy.spatial import Voronoi
from scipy.ndimage import gaussian_filter1d
from shapely.geometry import LineString, MultiLineString, Point, MultiPoint

from label_centerlines.exceptions import CenterlineError

logger = logging.getLogger(__name__)


def get_centerline(
    geom,
    segmentize_maxlen=0.5,
    max_points=3000,
    simplification=0.05,
    smooth_sigma=5,
    max_paths=5,
):
    """
    Return centerline from geometry.

    Parameters:
    -----------
    geom : shapely Polygon or MultiPolygon
    segmentize_maxlen : Maximum segment length for polygon borders.
        (default: 0.5)
    max_points : Number of points per geometry allowed before simplifying.
        (default: 3000)
    simplification : Simplification threshold.
        (default: 0.05)
    smooth_sigma : Smoothness of the output centerlines.
        (default: 5)
    max_paths : Number of longest paths used to create the centerlines.
        (default: 5)

    Returns:
    --------
    geometry : LineString or MultiLineString

    Raises:
    -------
    CenterlineError : if centerline cannot be extracted from Polygon
    TypeError : if input geometry is not Polygon or MultiPolygon

    """
    logger.debug("geometry type %s", geom.geom_type)

    if geom.geom_type == "Polygon":
        # segmentized Polygon outline
        outline = _segmentize(geom.exterior, segmentize_maxlen)
        logger.debug("outline: %s", outline)

        # simplify segmentized geometry if necessary and get points
        outline_points = outline.coords
        simplification_updated = simplification
        while len(outline_points) > max_points:
            # if geometry is too large, apply simplification until geometry
            # is simplified enough (indicated by the "max_points" value)
            simplification_updated += simplification
            outline_points = outline.simplify(simplification_updated).coords
        logger.debug("simplification used: %s", simplification_updated)
        logger.debug("simplified points: %s", MultiPoint(outline_points))

        # calculate Voronoi diagram and convert to graph but only use points
        # from within the original polygon
        vor = Voronoi(outline_points)
        graph = _graph_from_voronoi(vor, geom)
        logger.debug("voronoi diagram: %s", _multilinestring_from_voronoi(vor, geom))

        # determine longest path between all end nodes from graph
        end_nodes = _get_end_nodes(graph)
        if len(end_nodes) < 2:
            logger.debug("Polygon has too few points")
            raise CenterlineError("Polygon has too few points")
        logger.debug("get longest path from %s end nodes", len(end_nodes))
        longest_paths = _get_longest_paths(end_nodes, graph, max_paths)
        if not longest_paths:
            logger.debug("no paths found between end nodes")
            raise CenterlineError("no paths found between end nodes")
        if logger.getEffectiveLevel() <= 10:
            logger.debug("longest paths:")
            for path in longest_paths:
                logger.debug(LineString(vor.vertices[path]))

        # get least curved path from the longest paths, smooth and
        # return as LineString
        centerline = _smooth_linestring(
            LineString(
                vor.vertices[_get_least_curved_path(longest_paths, vor.vertices)]
            ),
            smooth_sigma,
        )
        logger.debug("centerline: %s", centerline)
        logger.debug("return linestring")
        return centerline

    elif geom.geom_type == "MultiPolygon":
        logger.debug("MultiPolygon found with %s sub-geometries", len(geom.geoms))
        # get centerline for each part Polygon and combine into MultiLineString
        sub_centerlines = []
        for subgeom in geom.geoms:
            try:
                sub_centerline = get_centerline(
                    subgeom, segmentize_maxlen, max_points, simplification, smooth_sigma
                )
                sub_centerlines.append(sub_centerline)
            except CenterlineError as e:
                logger.debug("subgeometry error: %s", e)
        # for MultPolygon, only raise CenterlineError if all subgeometries fail
        if sub_centerlines:
            return MultiLineString(sub_centerlines)
        else:
            raise CenterlineError("all subgeometries failed")

    else:
        raise TypeError(
            "Geometry type must be Polygon or MultiPolygon, not %s" % geom.geom_type
        )


# helper functions #
####################


def _segmentize(geom, max_len):
    """Interpolate points on segments if they exceed maximum length."""
    points = []
    for previous, current in zip(geom.coords, geom.coords[1:]):
        line_segment = LineString([previous, current])
        # add points on line segment if necessary
        points.extend(
            [
                line_segment.interpolate(max_len * i).coords[0]
                for i in range(int(line_segment.length / max_len))
            ]
        )
        # finally, add end point
        points.append(current)
    return LineString(points)


def _smooth_linestring(linestring, smooth_sigma):
    """Use a gauss filter to smooth out the LineString coordinates."""
    return LineString(
        zip(
            np.array(gaussian_filter1d(linestring.xy[0], smooth_sigma)),
            np.array(gaussian_filter1d(linestring.xy[1], smooth_sigma)),
        )
    )


def _get_longest_paths(nodes, graph, max_paths):
    """Return longest paths of all possible paths between a list of nodes."""

    def _gen_paths_distances():
        for node1, node2 in combinations(nodes, r=2):
            try:
                yield nx.single_source_dijkstra(
                    G=graph, source=node1, target=node2, weight="weight"
                )
            except NetworkXNoPath:
                continue

    return [x for (y, x) in sorted(_gen_paths_distances(), reverse=True)][:max_paths]


def _get_least_curved_path(paths, vertices):
    """Return path with smallest angles."""
    return min(
        zip([_get_path_angles_sum(path, vertices) for path in paths], paths),
        key=operator.itemgetter(0),
    )[1]


def _get_path_angles_sum(path, vertices):
    """Return all angles between edges from path."""
    return sum(
        [
            _get_absolute_angle(
                (vertices[pre], vertices[cur]), (vertices[cur], vertices[nex])
            )
            for pre, cur, nex in zip(path[:-1], path[1:], path[2:])
        ]
    )


def _get_absolute_angle(edge1, edge2):
    """Return absolute angle between edges."""
    v1 = edge1[0] - edge1[1]
    v2 = edge2[0] - edge2[1]
    return abs(np.degrees(np.math.atan2(np.linalg.det([v1, v2]), np.dot(v1, v2))))


def _get_end_nodes(graph):
    """Return list of nodes with just one neighbor node."""
    return [i for i in graph.nodes() if len(list(graph.neighbors(i))) == 1]


def _graph_from_voronoi(vor, geometry):
    """Return networkx.Graph from Voronoi diagram within geometry."""
    graph = nx.Graph()
    for x, y, dist in _yield_ridge_vertices(vor, geometry, dist=True):
        graph.add_nodes_from([x, y])
        graph.add_edge(x, y, weight=dist)
    return graph


def _multilinestring_from_voronoi(vor, geometry):
    """Return MultiLineString geometry from Voronoi diagram."""
    return MultiLineString(
        [
            LineString([Point(vor.vertices[[x, y]][0]), Point(vor.vertices[[x, y]][1])])
            for x, y in _yield_ridge_vertices(vor, geometry)
        ]
    )


def _yield_ridge_vertices(vor, geometry, dist=False):
    """Yield Voronoi ridge vertices within geometry."""
    for x, y in vor.ridge_vertices:
        if x < 0 or y < 0:
            continue
        point1 = Point(vor.vertices[[x, y]][0])
        point2 = Point(vor.vertices[[x, y]][1])
        # Eliminate all points outside our geometry.
        if point1.within(geometry) and point2.within(geometry):
            if dist:
                yield x, y, point1.distance(point2)
            else:
                yield x, y



# pmtiles to files
import gzip
import json
import os
import sqlite3
from pmtiles.writer import write
from pmtiles.reader import Reader, MmapSource, all_tiles
from .tile import zxy_to_tileid, tileid_to_zxy, TileType, Compression


def mbtiles_to_header_json(mbtiles_metadata):
    header = {}

    header["min_zoom"] = int(mbtiles_metadata["minzoom"])

    header["max_zoom"] = int(mbtiles_metadata["maxzoom"])

    bounds = mbtiles_metadata["bounds"].split(",")
    header["min_lon_e7"] = int(float(bounds[0]) * 10000000)
    header["min_lat_e7"] = int(float(bounds[1]) * 10000000)
    header["max_lon_e7"] = int(float(bounds[2]) * 10000000)
    header["max_lat_e7"] = int(float(bounds[3]) * 10000000)

    center = mbtiles_metadata["center"].split(",")
    header["center_lon_e7"] = int(float(center[0]) * 10000000)
    header["center_lat_e7"] = int(float(center[1]) * 10000000)
    header["center_zoom"] = int(center[2])

    tile_format = mbtiles_metadata["format"]
    if tile_format == "pbf":
        header["tile_type"] = TileType.MVT
    elif tile_format == "png":
        header["tile_type"] = TileType.PNG
    elif tile_format == "jpeg":
        header["tile_type"] = TileType.JPEG
    elif tile_format == "webp":
        header["tile_type"] = TileType.WEBP
    elif tile_format == "avif":
        header["tile_type"] = TileType.AVIF
    else:
        header["tile_type"] = TileType.UNKNOWN

    if tile_format == "pbf" or mbtiles_metadata.get("compression") == "gzip":
        header["tile_compression"] = Compression.GZIP
    else:
        header["tile_compression"] = Compression.NONE

    return header, mbtiles_metadata


def mbtiles_to_pmtiles(input, output, maxzoom):
    conn = sqlite3.connect(input)
    cursor = conn.cursor()

    with write(output) as writer:

        # collect a set of all tile IDs
        tileid_set = []
        for row in cursor.execute(
            "SELECT zoom_level,tile_column,tile_row FROM tiles WHERE zoom_level <= ?",
            (maxzoom or 99,),
        ):
            flipped = (1 << row[0]) - 1 - row[2]
            tileid_set.append(zxy_to_tileid(row[0], row[1], flipped))

        tileid_set.sort()

        mbtiles_metadata = {}
        for row in cursor.execute("SELECT name,value FROM metadata"):
            mbtiles_metadata[row[0]] = row[1]
        is_pbf = mbtiles_metadata["format"] == "pbf"

        # query the db in ascending tile order
        for tileid in tileid_set:
            z, x, y = tileid_to_zxy(tileid)
            flipped = (1 << z) - 1 - y
            res = cursor.execute(
                "SELECT tile_data FROM tiles WHERE zoom_level = ? AND tile_column = ? AND tile_row = ?",
                (z, x, flipped),
            )
            data = res.fetchone()[0]
            # force gzip compression only for vector
            if is_pbf and data[0:2] != b"\x1f\x8b":
                data = gzip.compress(data)
            writer.write_tile(tileid, data)

        pmtiles_header, pmtiles_metadata = mbtiles_to_header_json(mbtiles_metadata)
        if maxzoom:
            pmtiles_header["max_zoom"] = int(maxzoom)
            mbtiles_metadata["maxzoom"] = maxzoom
        result = writer.finalize(pmtiles_header, pmtiles_metadata)

    conn.close()


def pmtiles_to_mbtiles(input, output):
    conn = sqlite3.connect(output)
    cursor = conn.cursor()
    cursor.execute("CREATE TABLE metadata (name text, value text);")
    cursor.execute("""create unique index name on metadata (name);""")
    cursor.execute(
        "CREATE TABLE tiles (zoom_level integer, tile_column integer, tile_row integer, tile_data blob);"
    )
    cursor.execute(
        "CREATE UNIQUE INDEX tile_index on tiles (zoom_level, tile_column, tile_row);"
    )


    with open(input, "r+b") as f:
        source = MmapSource(f)

        reader = Reader(source)
        header = reader.header()
        metadata = reader.metadata()

        if "minzoom" not in metadata:
            metadata["minzoom"] = header["min_zoom"]

        if "maxzoom" not in metadata:
            metadata["maxzoom"] = header["max_zoom"]

        if "bounds" not in metadata:
            min_lon = header["min_lon_e7"] / 10000000
            min_lat = header["min_lat_e7"] / 10000000
            max_lon = header["max_lon_e7"] / 10000000
            max_lat = header["max_lat_e7"] / 10000000
            metadata["bounds"] = f"{min_lon},{min_lat},{max_lon},{max_lat}"

        if "center" not in metadata:
            center_lon = header["center_lon_e7"] / 10000000
            center_lat = header["center_lat_e7"] / 10000000
            center_zoom = header["center_zoom"]
            metadata["center"] = f"{center_lon},{center_lat},{center_zoom}"

        if "format" not in metadata:
            if header["tile_type"] == TileType.MVT:
                metadata["format"] = "pbf"

        json_metadata = {}
        for k, v in metadata.items():
            if k == "vector_layers":
                json_metadata["vector_layers"] = v
                continue
            elif k == "tilestats":
                json_metadata["tilestats"] = v
                continue
            elif not isinstance(v, str):
                v = json.dumps(v, ensure_ascii=False)
            cursor.execute("INSERT INTO metadata VALUES(?,?)", (k, v))

        if len(json_metadata) > 0:
            cursor.execute(
                "INSERT INTO metadata VALUES(?,?)",
                ("json", json.dumps(json_metadata, ensure_ascii=False)),
            )

        for zxy, tile_data in all_tiles(source):
            flipped_y = (1 << zxy[0]) - 1 - zxy[2]
            cursor.execute(
                "INSERT INTO tiles VALUES(?,?,?,?)",
                (zxy[0], zxy[1], flipped_y, tile_data),
            )

    conn.commit()
    conn.close()


def pmtiles_to_dir(input, output):
    os.makedirs(output)
    with open(input, "r+b") as f:
        source = MmapSource(f)

        reader = Reader(source)
        with open(os.path.join(output, "metadata.json"), "w") as f:
            f.write(json.dumps(reader.metadata()))

        for zxy, tile_data in all_tiles(source):
            directory = os.path.join(output, str(zxy[0]), str(zxy[1]))
            path = os.path.join(directory, str(zxy[2]) + "." + "mvt")
            os.makedirs(directory, exist_ok=True)
            with open(path, "wb") as f:
                f.write(tile_data)



#!/usr/bin/env python
import os
import shutil
import sys
import argparse
from tqdm import tqdm

def convert_tms_to_xyz(inDIR, copyDIR):
    # Check if output directory exists, exit if it does
    if os.path.exists(copyDIR):
        sys.exit("ERROR: The output directory already exists!")

    # Create the root of the output directory if it does not exist
    if not os.path.exists(copyDIR):
        os.makedirs(copyDIR)

    # Copy all files from the root of inDIR to the root of copyDIR, including metadata.json
    root_files = [f for f in os.listdir(inDIR) if os.path.isfile(os.path.join(inDIR, f))]
    for root_file in tqdm(root_files, desc="Copying metadata (if existed)", unit=' ', ncols=80, bar_format='{l_bar}{bar}| {n_fmt}/{total_fmt} [{percentage:.0f}%]'):
        shutil.copy(os.path.join(inDIR, root_file), os.path.join(copyDIR, root_file))

    # Count the total number of files to process
    total_files = sum([len(files) for rVal, dName, files in os.walk(inDIR) if rVal != inDIR])

    # Initialize the progress bar for nested files
    with tqdm(total=total_files, desc="Processing files", unit=' ', ncols=80, bar_format='{l_bar}{bar}| {n_fmt}/{total_fmt} [{percentage:.0f}%]') as pbar:
        for rVal, dName, fList in os.walk(inDIR):
            # Skip the root directory as it has already been handled
            if rVal == inDIR:
                continue

            # Calculate the relative path for the new directory structure
            relPath = os.path.relpath(rVal, inDIR)
            newpath = os.path.join(copyDIR, relPath)
            if not os.path.exists(newpath):
                os.makedirs(newpath)

            for fileName in fList:
                zxParts = os.path.normpath(relPath).split(os.sep)
                if len(zxParts) < 2:
                    print(f"Skipping file due to insufficient path depth: {fileName}")
                    continue

                z = int(zxParts[0])
                y = int(fileName.split(".")[0])
                fileExtension = fileName.split(".")[1]
                newY = str((2**z - 1) - y) + "." + fileExtension
                shutil.copyfile(os.path.join(rVal, fileName), os.path.join(newpath, newY))
                pbar.update(1)

def main():
    # Get the command line arguments and set up some variables
    parser = argparse.ArgumentParser(description='Convert TMS <--> XYZ tiling scheme for a tiles folder')
    parser.add_argument('-i', type=str, help='input directorty')
    parser.add_argument('-o', type=str, help='output directory')
    args = parser.parse_args()
    
    convert_tms_to_xyz(args.i, args.o)

if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# Reference: https://github.com/monkut/servembtiles
"""
mbtiles WSGI application

MBTiles is a specification for storing tiled map data in SQLite databases for immediate usage and for transfer.
From:
https://github.com/mapbox/mbtiles-spec
"""
import os
import json
import sqlite3
import mimetypes
import logging
from wsgiref.util import shift_path_info
from wsgiref.simple_server import make_server, WSGIServer
from socketserver import ThreadingMixIn


logger = logging.getLogger(__name__)
logger.setLevel(logging.WARNING)

try:
    from settings import MBTILES_ABSPATH, MBTILES_TILE_EXT, MBTILES_ZOOM_OFFSET, MBTILES_HOST, MBTILES_PORT, MBTILES_SERVE, USE_OSGEO_TMS_TILE_ADDRESSING
except ImportError:
    logger.warning("settings.py not set, may not be able to run via a web server (apache, nginx, etc)!")
    MBTILES_ABSPATH = None
    MBTILES_TILE_EXT = '.png'
    MBTILES_ZOOM_OFFSET = 0
    MBTILES_HOST = ''
    MBTILES_PORT = 8005
    MBTILES_SERVE = False
    USE_OSGEO_TMS_TILE_ADDRESSING = True

SUPPORTED_IMAGE_EXTENSIONS = (".png", ".jpg", ".jpeg")


class MBTilesFileNotFound(Exception):
    pass


class InvalidImageExtension(Exception):
    pass


class MBTilesApplication:
    """
    Serves rendered tiles within the given .mbtiles (sqlite3) file defined in settings.MBTILES_ABSPATH

    Refer to the MBTiles specification at:
    https://github.com/mapbox/mbtiles-spec
    """

    def __init__(self, mbtiles_filepath, tile_image_ext='.png', zoom_offset=0):
        if mbtiles_filepath is None or not os.path.exists(mbtiles_filepath):
            raise MBTilesFileNotFound(mbtiles_filepath)

        if tile_image_ext not in SUPPORTED_IMAGE_EXTENSIONS:
            raise InvalidImageExtension("{} not in {}!".format(tile_image_ext, SUPPORTED_IMAGE_EXTENSIONS))

        self.mbtiles_db = sqlite3.connect(
            "file:{}?mode=ro".format(mbtiles_filepath),
            check_same_thread=False, uri=True)
        self.tile_image_ext = tile_image_ext
        self.tile_content_type = mimetypes.types_map[tile_image_ext.lower()]
        self.zoom_offset = zoom_offset
        self.maxzoom = None
        self.minzoom = None

        self._populate_supported_zoom_levels()

    def _populate_supported_zoom_levels(self):
        """
        Query the metadata table and obtain max/min zoom levels,
        setting to self.minzoom, self.maxzoom as integers
        :return: None
        """
        query = 'SELECT name, value FROM metadata WHERE name="minzoom" OR name="maxzoom";'
        # add maxzoom, minzoom to instance
        for name, value in self.mbtiles_db.execute(query):
            setattr(self, name.lower(), max(int(value) - self.zoom_offset, 0))

    def __call__(self, environ, start_response):
        if environ['REQUEST_METHOD'] == 'GET':
            uri_field_count = len(environ['PATH_INFO'].split('/'))
            base_uri = shift_path_info(environ)

            # handle 'metadata' requests
            if base_uri == 'metadata':
                query = 'SELECT * FROM metadata;'
                metadata_results = self.mbtiles_db.execute(query).fetchall()
                if metadata_results:
                    status = '200 OK'
                    response_headers = [('Content-type', 'application/json')]
                    start_response(status, response_headers)
                    json_result = json.dumps(metadata_results, ensure_ascii=False)
                    return [json_result.encode("utf8"),]
                else:
                    status = '404 NOT FOUND'
                    response_headers = [('Content-type', 'text/plain; charset=utf-8')]
                    start_response(status, response_headers)
                    return ['"metadata" not found in configured .mbtiles file!'.encode('utf8'), ]

            # handle tile request
            elif uri_field_count >= 3:  # expect:  zoom, x & y
                try:
                    zoom = int(base_uri)
                    if None not in (self.minzoom, self.maxzoom) and not (self.minzoom <= zoom <= self.maxzoom):
                        status = "404 Not Found"
                        response_headers = [('Content-type', 'text/plain; charset=utf-8')]
                        start_response(status, response_headers)
                        return ['Requested zoomlevel({}) Not Available! Valid range minzoom({}) maxzoom({}) PATH_INFO: {}'.format(zoom,
                                                                                                                                   self.minzoom,
                                                                                                                                   self.maxzoom,
                                                                                                                                   environ['PATH_INFO']).encode('utf8')]
                    zoom += self.zoom_offset
                    x = int(shift_path_info(environ))
                    y, ext = shift_path_info(environ).split('.')
                    y = int(y)
                except ValueError as e:
                    status = "400 Bad Request"
                    response_headers = [('Content-type', 'text/plain; charset=utf-8')]
                    start_response(status, response_headers)
                    return ['Unable to parse PATH_INFO({}), expecting "z/x/y.(png|jpg)"'.format(environ['PATH_INFO']).encode('utf8'), ' '.join(i for i in e.args).encode('utf8')]

                query = 'SELECT tile_data FROM tiles WHERE zoom_level=? AND tile_column=? AND tile_row=?;'
                if not USE_OSGEO_TMS_TILE_ADDRESSING:
                    # adjust y to use XYZ google addressing
                    ymax = 1 << zoom
                    y = ymax - y - 1
                values = (zoom, x, y)
                tile_results = self.mbtiles_db.execute(query, values).fetchone()

                if tile_results:
                    tile_result = tile_results[0]
                    status = '200 OK'
                    response_headers = [('Content-type', self.tile_content_type)]
                    start_response(status, response_headers)
                    return [tile_result,]
                else:
                    status = '404 NOT FOUND'
                    response_headers = [('Content-type', 'text/plain; charset=utf-8')]
                    start_response(status, response_headers)
                    return ['No data found for request location: {}'.format(environ['PATH_INFO']).encode('utf8')]

        status = "400 Bad Request"
        response_headers = [('Content-type', 'text/plain; charset=utf-8')]
        start_response(status, response_headers)
        return ['request URI not in expected: ("metadata", "/z/x/y.png")'.encode('utf8'), ]


def main():
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument("--serve",
                        default=MBTILES_SERVE,
                        action='store_true',
                        help="Start test server [DEFAULT={}]\n(Defaults to environment variable, 'MBTILES_SERVE')".format(MBTILES_SERVE))
    parser.add_argument('-p', '--port',
                        default=MBTILES_PORT,
                        type=int,
                        help="Test server port [DEFAULT={}]\n(Defaults to environment variable, 'MBTILES_PORT')".format(MBTILES_PORT))
    parser.add_argument('-a', '--address',
                        default=MBTILES_HOST,
                        help="Test address to serve on [DEFAULT=\"{}\"]\n(Defaults to environment variable, 'MBTILES_HOST')".format(MBTILES_HOST))
    parser.add_argument('-f', '--filepath',
                        default=MBTILES_ABSPATH,
                        help="mbtiles filepath [DEFAULT={}]\n(Defaults to environment variable, 'MBTILES_ABSPATH')".format(MBTILES_ABSPATH))
    parser.add_argument('-e', '--ext',
                        default=MBTILES_TILE_EXT,
                        help="mbtiles image file extension [DEFAULT={}]\n(Defaults to environment variable, 'MBTILES_TILE_EXT')".format(MBTILES_TILE_EXT))
    parser.add_argument('-z', '--zoom-offset',
                        default=MBTILES_ZOOM_OFFSET,
                        type=int,
                        help="mbtiles zoom offset [DEFAULT={}]\n(Defaults to environment variable, 'MBTILES_ZOOM_OFFSET')".format(MBTILES_ZOOM_OFFSET))
    args = parser.parse_args()
    args.filepath = os.path.abspath(args.filepath)

    if args.serve:
        # create console handler and set level to debug
        console = logging.StreamHandler()
        console.setLevel(logging.DEBUG)
        formatter = logging.Formatter('%(asctime)s - %(module)s - %(levelname)s - %(message)s')
        console.setFormatter(formatter)
        logger.addHandler(console)
        logger.setLevel(logging.DEBUG)

        logger.info("FILEPATH: {}".format(args.filepath))
        logger.info("TILE EXT: {}".format(args.ext))
        logger.info("ADDRESS : {}".format(args.address))
        logger.info("PORT    : {}".format(args.port))

        class ThreadingWSGIServer(ThreadingMixIn, WSGIServer): pass

        mbtiles_app = MBTilesApplication(mbtiles_filepath=args.filepath, tile_image_ext=args.ext, zoom_offset=args.zoom_offset)
        server = make_server(args.address, args.port, mbtiles_app, ThreadingWSGIServer)
        try:
            server.serve_forever()
        except KeyboardInterrupt:
            logger.info("stopped.")
    else:
        logger.warning("'--serve' option not given!")
        logger.warning("\tRun with the '--serve' option to serve tiles with the test server.")


if __name__ == '__main__':
    main()


import json
from collections import namedtuple


def load_directory(data_bytes, offset, num_entries):
    tile_entries = {}
    leaves = {}
    for i in range(offset, offset + num_entries * 17, 17):
        z = int.from_bytes(data_bytes[i : i + 1], byteorder="little")
        x = int.from_bytes(data_bytes[i + 1 : i + 4], byteorder="little")
        y = int.from_bytes(data_bytes[i + 4 : i + 7], byteorder="little")
        tile_off = int.from_bytes(data_bytes[i + 7 : i + 13], byteorder="little")
        tile_len = int.from_bytes(data_bytes[i + 13 : i + 17], byteorder="little")
        if z & 0b10000000:
            leaves[(z & 0b01111111, x, y)] = (tile_off, tile_len)
        else:
            tile_entries[(z, x, y)] = (tile_off, tile_len)
    return tile_entries, leaves


Header = namedtuple("Header", ["version", "metadata", "root_dir", "leaves"])


class Reader:
    def __init__(self, get_bytes):
        self.get_bytes = get_bytes
        self._header = None

    def header(self):
        if self._header:
            return self._header
        else:
            header_bytes = self.get_bytes(0, 512000)
            assert int.from_bytes(header_bytes[0:2], byteorder="little") == 0x4D50
            version = int.from_bytes(header_bytes[2:4], byteorder="little")
            metadata_len = int.from_bytes(header_bytes[4:8], byteorder="little")
            metadata = json.loads(header_bytes[10 : 10 + metadata_len])
            num_entries = int.from_bytes(header_bytes[8:10], byteorder="little")
            root_dir, leaves = load_directory(
                header_bytes, 10 + metadata_len, num_entries
            )
            self._header = Header(version, metadata, root_dir, leaves)
            return self._header

    def _leaf_level(self):
        h = self.header()
        return next(iter(h.leaves))[0]

    def get(self, z, x, y):
        h = self.header()
        val = h.root_dir.get((z, x, y))
        if val:
            return self.get_bytes(val[0], val[1])
        else:
            if len(self.header().leaves) > 0:
                level_diff = z - self._leaf_level()
                if level_diff < 0:
                    return None
                leaf = (
                    self._leaf_level(),
                    x // (1 << level_diff),
                    y // (1 << level_diff),
                )
                val = h.leaves.get(leaf)
                if val:
                    dir_bytes = self.get_bytes(val[0], val[1])
                    directory, _ = load_directory(dir_bytes, 0, val[1] // 17)
                    val = directory.get((z, x, y))
                    if val:
                        return self.get_bytes(val[0], val[1])

    def tiles(self):
        h = self.header()
        for k, v in h.root_dir.items():
            yield (k, self.get_bytes(v[0], v[1]))
        for val in set(h.leaves.values()):
            dir_bytes = self.get_bytes(val[0], val[1])
            leaf_dir, _ = load_directory(dir_bytes, 0, val[1] // 17)
            for k, v in leaf_dir.items():
                yield (k, self.get_bytes(v[0], v[1]))



import logging
from datetime import datetime

from geojson2vt.convert import convert
from geojson2vt.clip import clip
from geojson2vt.wrap import wrap
from geojson2vt.transform import transform_tile
from geojson2vt.tile import create_tile


def get_default_options():
    return {
        "maxZoom": 14,            # max zoom to preserve detail on
        "indexMaxZoom": 5,        # max zoom in the tile index
        "indexMaxPoints": 100000,  # max number of points per tile in the tile index
        # simplification tolerance (higher means simpler)
        "tolerance": 3,
        "extent": 4096,           # tile extent
        "buffer": 64,             # tile buffer on each side
        "lineMetrics": False,     # whether to calculate line metrics
        "promoteId": None,        # name of a feature property to be promoted to feature.id
        "generateId": False,      # whether to generate feature ids. Cannot be used with promoteId
    }


class GeoJsonVt:
    def __init__(self, data, options, log_level=logging.INFO):
        logging.basicConfig(
            level=log_level, format='%(asctime)s %(levelname)s %(message)s')
        options = self.options = extend(get_default_options(), options)

        logging.debug('preprocess data start')

        if options.get('maxZoom') < 0 or options.get('maxZoom') > 24:
            raise Exception('maxZoom should be in the 0-24 range')
        if options.get('promoteId', None) is not None and options.get('generateId', False):
            raise Exception(
                'promoteId and generateId cannot be used together.')

        # projects and adds simplification info
        features = convert(data, options)

        # tiles and tile_coords are part of the public API
        self.tiles = {}
        self.tile_coords = []

        logging.debug(f'preprocess data end')
        logging.debug(
            f'index: maxZoom: {options.get("indexMaxZoom")}, maxPoints: {options.get("indexMaxPoints")}')
        self.stats = {}
        self.total = 0

        # wraps features (ie extreme west and extreme east)
        features = wrap(features, options)

        # start slicing from the top tile down
        if len(features) > 0:
            self.split_tile(features, 0, 0, 0)

        if len(features) > 0:
            logging.debug(
                f'features: {self.tiles[0].get("numFeatures")}, points: {self.tiles[0].get("numPoints")}')
            stop = datetime.now()
        logging.debug(f'generate tiles end')
        logging.debug('tiles generated:', self.total, self.stats)

    # splits features from a parent tile to sub-tiles.
    # z, x, and y are the coordinates of the parent tile
    # cz, cx, and cy are the coordinates of the target tile
    #
    # If no target tile is specified, splitting stops when we reach the maximum
    # zoom or the number of points is low as specified in the options.

    def split_tile(self, features, z, x, y, cz=None, cx=None, cy=None):
        stack = [features, z, x, y]
        options = self.options
        # avoid recursion by using a processing queue
        while len(stack) > 0:
            y = stack.pop()
            x = stack.pop()
            z = stack.pop()
            features = stack.pop()

            z2 = 1 << z
            id_ = to_Id(z, x, y)
            tile = self.tiles.get(id_, None)

            if tile is None:
                logging.debug('creation start')

                self.tiles[id_] = create_tile(features, z, x, y, options)
                tile = self.tiles[id_]
                self.tile_coords.append({'z': z, 'x': x, 'y': y})

                logging.debug(
                    f'tile z{z}-{x}-{y} (features: {tile.get("numFeatures")}, points: {tile.get("numPoints")}, simplified: {tile.get("numSimplified")})')
                logging.debug(f'creation end')
                key = f'z{z}'
                self.stats[key] = self.stats.get(key, 0) + 1
                self.total += 1

            # save reference to original geometry in tile so that we can drill down later if we stop now
            tile['source'] = features

            # if it's the first-pass tiling
            if cz is None:
                # stop tiling if we reached max zoom, or if the tile is too simple
                if z == options.get('indexMaxZoom') or tile.get('numPoints') <= options.get('indexMaxPoints'):
                    continue  # if a drilldown to a specific tile
            elif z == options.get('maxZoom') or z == cz:
                # stop tiling if we reached base zoom or our target tile zoom
                continue
            elif cz is not None:
                # stop tiling if it's not an ancestor of the target tile
                zoomSteps = cz - z
                if x != (cx >> zoomSteps) or y != (cy >> zoomSteps):
                    continue

            # if we slice further down, no need to keep source geometry
            tile['source'] = None

            if not features or len(features) == 0:
                continue

            logging.debug('clipping start')

            # values we'll use for clipping
            k1 = 0.5 * options.get('buffer') / options.get('extent')
            k2 = 0.5 - k1
            k3 = 0.5 + k1
            k4 = 1 + k1

            tl = None
            bl = None
            tr = None
            br = None

            left = clip(features, z2, x - k1, x + k3, 0,
                        tile['minX'], tile['maxX'], options)
            right = clip(features, z2, x + k2, x + k4, 0,
                         tile['minX'], tile['maxX'], options)
            features = None

            if left is not None:
                tl = clip(left, z2, y - k1, y + k3, 1,
                          tile['minY'], tile['maxY'], options)
                bl = clip(left, z2, y + k2, y + k4, 1,
                          tile['minY'], tile['maxY'], options)
                left = None

            if right is not None:
                tr = clip(right, z2, y - k1, y + k3, 1,
                          tile['minY'], tile['maxY'], options)
                br = clip(right, z2, y + k2, y + k4, 1,
                          tile['minY'], tile['maxY'], options)
                right = None

            logging.debug(f'clipping took end')

            #stack.push(tl or [], z + 1, x * 2,     y * 2)
            stack.append(tl if tl is not None else [])
            stack.append(z + 1)
            stack.append(x * 2)
            stack.append(y * 2)

            #stack.push(bl or [], z + 1, x * 2,     y * 2 + 1)
            stack.append(bl if bl is not None else [])
            stack.append(z + 1)
            stack.append(x * 2)
            stack.append(y * 2 + 1)

            #stack.push(tr or [], z + 1, x * 2 + 1, y * 2)
            stack.append(tr if tr is not None else [])
            stack.append(z + 1)
            stack.append(x * 2 + 1)
            stack.append(y * 2)

            #stack.push(br or [], z + 1, x * 2 + 1, y * 2 + 1)
            stack.append(br if br is not None else [])
            stack.append(z + 1)
            stack.append(x * 2 + 1)
            stack.append(y * 2 + 1)

    def get_tile(self, z, x, y):
        z = int(z)
        x = int(x)
        y = int(y)

        options = self.options
        extent = options.get('extent')

        if z < 0 or z > 24:
            return None

        z2 = 1 << z
        x = (x + z2) & (z2 - 1)  # wrap tile x coordinate

        id_ = to_Id(z, x, y)
        current_tile = self.tiles.get(id_, None)
        if current_tile is not None:
            return transform_tile(self.tiles[id_], extent)

        logging.debug(f'drilling down to z{z}-{x}-{y}')

        z0 = z
        x0 = x
        y0 = y
        parent = None

        while parent is None and z0 > 0:
            z0 -= 1
            x0 = x0 >> 1
            y0 = y0 >> 1
            parent = self.tiles.get(to_Id(z0, x0, y0), None)

        if parent is None or parent.get('source', None) is None:
            return None

        # if we found a parent tile containing the original geometry, we can drill down from it
        logging.debug(f'found parent tile z{z0}-{x0}-{y0}')
        logging.debug('drilling down start')

        self.split_tile(parent.get('source'), z0, x0, y0, z, x, y)

        logging.debug(f'drilling down end')

        transformed = transform_tile(self.tiles[id_], extent) if self.tiles.get(
            id_, None) is not None else None
        return transformed


def to_Id(z, x, y):
    id_ = (((1 << z) * y + x) * 32) + z
    return id_


def extend(dest, src):
    for key, _ in src.items():
        dest[key] = src[key]
    return dest


def geojson2vt(data, options, log_level=logging.INFO):
    return GeoJsonVt(data, options, log_level)



import sqlite3, json
import os, sys, datetime
from vtiles.utils.geopreocessing import check_vector,count_tiles

# Read vector metadata
def read_vector_metadata(mbtiles):
    try:
        """Read metadata from vector MBTiles file."""
        connection = sqlite3.connect(mbtiles)
        cursor = connection.cursor()
        # Extract metadata
        # cursor.execute("SELECT name, value FROM metadata where name <>'json'")
        cursor.execute("SELECT name, value FROM metadata where name <> 'json'")
        metadata = dict(cursor.fetchall())     
        return metadata
    except sqlite3.Error as e:
            print(f"error reading metadata: {e}")
            print(f"Please use mbtilesfixmeta to create metadata or use mbtilesinspect")
            return None
    finally:
        cursor.close()
        connection.close()  

# Read raster metadata
def read_raster_metadata(mbtiles):
    try:
        """Read metadata from raster MBTiles file."""
        connection = sqlite3.connect(mbtiles)
        cursor = connection.cursor()
        # Extract metadata
        cursor.execute("SELECT name, value FROM metadata")
        metadata = dict(cursor.fetchall())
        return metadata
    except sqlite3.Error as e:
        print(f"error reading metadata: {e}")
        print(f"Please use mbtilesfixmeta to create metadata or use mbtilesinspect")
        return None
    finally:
        cursor.close()
        connection.close() 

# list all vector layers
def read_vector_layers(mbtiles):    
    connection = sqlite3.connect(mbtiles)
    cursor = connection.cursor()
    
    # Fetch the JSON from the metadata
    cursor.execute("SELECT value FROM metadata WHERE name = 'json'")
    row = cursor.fetchone()
    
    if row is not None:
        json_content = row[0]
        layers_json = json.loads(json_content)
        
        # Print vector layers information
        if "vector_layers" in layers_json:
            vector_layers = layers_json["vector_layers"]
            print("######:")
            print("Vector layers:")
            for index, layer in enumerate(vector_layers):
                row_index = index + 1
                layer_id = layer["id"]
                description = layer.get("description", "")
                minzoom = layer.get("minzoom", "")
                maxzoom = layer.get("maxzoom", "")
                fields = layer.get("fields", {})
                
                print(f"{row_index}: {layer_id}")
                # print(f"  Description: {description}")
                # print(f"  Min Zoom: {minzoom}")
                # print(f"  Max Zoom: {maxzoom}")
                # print(f"  Fields: {fields}")
                # print(" ")
        
        # Print tilestats information
        # if "tilestats" in layers_json:
        #     tilestats = layers_json["tilestats"]
        #     print("######:")
        #     print("Tile Stats:")
        #     layer_count = tilestats.get("layerCount", 0)
        #     print(f"  Layer Count: {layer_count}")
            
        #     if "layers" in tilestats:
        #         for index, layer in enumerate(tilestats["layers"]):
        #             row_index = index + 1
        #             layer_name = layer.get("layer", "")
        #             count = layer.get("count", "")
        #             geometry = layer.get("geometry", "")
        #             attribute_count = layer.get("attributeCount", 0)
                    
        #             print(f"{row_index}: {layer_name}")
        #             print(f"  Count: {count}")
        #             print(f"  Geometry: {geometry}")
        #             print(f"  Attribute Count: {attribute_count}")
                    
        #             if "attributes" in layer:
        #                 for attr_index, attribute in enumerate(layer["attributes"]):
        #                     attr_name = attribute.get("attribute", "")
        #                     attr_count = attribute.get("count", 0)
        #                     attr_type = attribute.get("type", "")
        #                     attr_values = attribute.get("values", [])
                            
        #                     print(f"    Attribute {attr_index + 1}: {attr_name}")
        #                     print(f"      Count: {attr_count}")
        #                     print(f"      Type: {attr_type}")
        #                     print(f"      Values: {attr_values}")
        #                     print(" ")            
    cursor.close()
    connection.close()
   
def main():
    if len(sys.argv) != 2:
      print("Please provide MBTiles input filename.")
      return
    mbtiles = sys.argv[1]
    file_stat = os.stat(mbtiles)
    file_size_mb = round(file_stat.st_size / (1024 * 1024),2)
    # file_created = datetime.datetime.fromtimestamp(file_stat.st_ctime).strftime("%Y-%m-%d %I:%M:%S %p")
    file_last_modified = datetime.datetime.fromtimestamp(file_stat.st_mtime).strftime("%Y-%m-%d %I:%M:%S %p")
    # file_size = os.path.getsize(mbtiles)
    # file_size_mb = round(file_size / (1024 * 1024),2)
    print('######')
    print("File size: ", file_size_mb, 'MB')
    # print("Date created: ", file_created)
    print("Last modified: ", file_last_modified)
    if (os.path.exists(mbtiles)):
        is_vector, _ = check_vector(mbtiles) 
        num_tiles = count_tiles(mbtiles)
        if is_vector: # vector
            metadata = read_vector_metadata(mbtiles)
            if metadata:      
                print("######")
                print("Metadata:")
                for key, value in metadata.items():
                    print(f"{key}: {value}")
                read_vector_layers(mbtiles)
            print('######')
            print(f"Total number of tiles: {num_tiles}")
            
        else:
            metadata = read_raster_metadata(mbtiles)   
            if metadata:         
                print("###### Metadata:")
                for key, value in metadata.items():
                    print(f"{key}: {value}")
            print("######")
            print(f"Total number of tiles: {num_tiles}")
    else: 
        print ('MBTiles file does not exist!. Please recheck and input a correct file path.')
        return

if __name__ == "__main__":
    main()



from urllib.parse import urlparse

from vtiles.utils.mapbox_vector_tile import decode

from .features import Layer


def _is_url(uri: str) -> bool:
    """
    Checks if the uri is actually an url.
    :param uri: the string to check.
    :return: a boolean.
    """
    return urlparse(uri).scheme != ""


def vt_bytes_to_geojson(b_content: bytes, x: int, y: int, z: int, layer=None) -> dict:
    """
    Make a GeoJSON from bytes in the vector tiles format.
    :param b_content: the bytes to convert.
    :param x: tile x coordinate.
    :param y: tile y coordinate.
    :param z: tile z coordinate.
    :param layer: include only the specified layer.
    :return: a features collection (GeoJSON).
    """
    data = decode(b_content, y_coord_down=True)

    features_collections = [Layer(x=x, y=y, z=z, name=layer_name, obj=layer_obj).toGeoJSON()
                            for layer_name, layer_obj in data.items() if layer is None or layer_name == layer]

    geojson = {fc["name"]: {"type": fc["type"], "features": fc["features"]} for fc in features_collections}

     
    return geojson
    



from enum import Enum
import io
import gzip


class Entry:
    __slots__ = ("tile_id", "offset", "length", "run_length")

    def __init__(self, tile_id, offset, length, run_length):
        self.tile_id = tile_id
        self.offset = offset
        self.length = length
        self.run_length = run_length

    def __str__(self):
        return f"id={self.tile_id} offset={self.offset} length={self.length} runlength={self.run_length}"


def rotate(n, xy, rx, ry):
    if ry == 0:
        if rx == 1:
            xy[0] = n - 1 - xy[0]
            xy[1] = n - 1 - xy[1]
        xy[0], xy[1] = xy[1], xy[0]


def t_on_level(z, pos):
    n = 1 << z
    rx, ry, t = pos, pos, pos
    xy = [0, 0]
    s = 1
    while s < n:
        rx = 1 & (t // 2)
        ry = 1 & (t ^ rx)
        rotate(s, xy, rx, ry)
        xy[0] += s * rx
        xy[1] += s * ry
        t //= 4
        s *= 2
    return z, xy[0], xy[1]


def zxy_to_tileid(z, x, y):
    if z > 31:
        raise OverflowError("tile zoom exceeds 64-bit limit")
    if x > (1 << z) - 1 or y > (1 << z) - 1:
        raise ValueError("tile x/y outside zoom level bounds")
    acc = 0
    tz = 0
    while tz < z:
        acc += (0x1 << tz) * (0x1 << tz)
        tz += 1
    n = 1 << z
    rx = 0
    ry = 0
    d = 0
    xy = [x, y]
    s = n // 2
    while s > 0:
        if (xy[0] & s) > 0:
            rx = 1
        else:
            rx = 0
        if (xy[1] & s) > 0:
            ry = 1
        else:
            ry = 0
        d += s * s * ((3 * rx) ^ ry)
        rotate(s, xy, rx, ry)
        s //= 2
    return acc + d


def tileid_to_zxy(tile_id):
    num_tiles = 0
    acc = 0
    for z in range(0,32):
        num_tiles = (1 << z) * (1 << z)
        if acc + num_tiles > tile_id:
            return t_on_level(z, tile_id - acc)
        acc += num_tiles
    raise OverflowError("tile zoom exceeds 64-bit limit")


def find_tile(entries, tile_id):
    m = 0
    n = len(entries) - 1
    while m <= n:
        k = (n + m) >> 1
        c = tile_id - entries[k].tile_id
        if c > 0:
            m = k + 1
        elif c < 0:
            n = k - 1
        else:
            return entries[k]

    if n >= 0:
        if entries[n].run_length == 0:
            return entries[n]
        if tile_id - entries[n].tile_id < entries[n].run_length:
            return entries[n]


def read_varint(b_io):
    shift = 0
    result = 0
    while True:
        raw = b_io.read(1)
        if raw == b"":
            raise EOFError("unexpectedly reached end of varint stream")
        i = ord(raw)
        result |= (i & 0x7F) << shift
        shift += 7
        if not (i & 0x80):
            break
    return result


def write_varint(b_io, i):
    while True:
        towrite = i & 0x7F
        i >>= 7
        if i:
            b_io.write(bytes([towrite | 0x80]))
        else:
            b_io.write(bytes([towrite]))
            break


class Compression(Enum):
    UNKNOWN = 0
    NONE = 1
    GZIP = 2
    BROTLI = 3
    ZSTD = 4


class TileType(Enum):
    UNKNOWN = 0
    MVT = 1
    PNG = 2
    JPEG = 3
    WEBP = 4
    AVIF = 5


def deserialize_directory(buf):
    b_io = io.BytesIO(gzip.decompress(buf))
    entries = []
    num_entries = read_varint(b_io)

    last_id = 0
    for i in range(num_entries):
        tmp = read_varint(b_io)
        entries.append(Entry(last_id + tmp, 0, 0, 0))
        last_id += tmp

    for i in range(num_entries):
        entries[i].run_length = read_varint(b_io)

    for i in range(num_entries):
        entries[i].length = read_varint(b_io)

    for i in range(num_entries):
        tmp = read_varint(b_io)
        if i > 0 and tmp == 0:
            entries[i].offset = entries[i - 1].offset + entries[i - 1].length
        else:
            entries[i].offset = tmp - 1

    return entries


def serialize_directory(entries):
    b_io = io.BytesIO()
    write_varint(b_io, len(entries))

    last_id = 0
    for e in entries:
        write_varint(b_io, e.tile_id - last_id)
        last_id = e.tile_id

    for e in entries:
        write_varint(b_io, e.run_length)

    for e in entries:
        write_varint(b_io, e.length)

    for i, e in enumerate(entries):
        if i > 0 and e.offset == entries[i - 1].offset + entries[i - 1].length:
            write_varint(b_io, 0)
        else:
            write_varint(b_io, e.offset + 1)

    return gzip.compress(b_io.getvalue())

class SpecVersionUnsupported(Exception):
    pass

class MagicNumberNotFound(Exception):
    pass

def deserialize_header(buf):
    if buf[0:7].decode() != "PMTiles":
        raise MagicNumberNotFound()

    if buf[7] != 0x3:
        raise SpecVersionUnsupported()

    def read_uint64(pos):
        return int.from_bytes(buf[pos : pos + 8], byteorder="little")

    def read_int32(pos):
        return int.from_bytes(buf[pos : pos + 4], byteorder="little", signed=True)

    return {
        "version": buf[7],
        "root_offset": read_uint64(8),
        "root_length": read_uint64(16),
        "metadata_offset": read_uint64(24),
        "metadata_length": read_uint64(32),
        "leaf_directory_offset": read_uint64(40),
        "leaf_directory_length": read_uint64(48),
        "tile_data_offset": read_uint64(56),
        "tile_data_length": read_uint64(64),
        "addressed_tiles_count": read_uint64(72),
        "tile_entries_count": read_uint64(80),
        "tile_contents_count": read_uint64(88),
        "clustered": buf[96] == 0x1,
        "internal_compression": Compression(buf[97]),
        "tile_compression": Compression(buf[98]),
        "tile_type": TileType(buf[99]),
        "min_zoom": buf[100],
        "max_zoom": buf[101],
        "min_lon_e7": read_int32(102),
        "min_lat_e7": read_int32(106),
        "max_lon_e7": read_int32(110),
        "max_lat_e7": read_int32(114),
        "center_zoom": buf[118],
        "center_lon_e7": read_int32(119),
        "center_lat_e7": read_int32(123),
    }


def serialize_header(h):
    b_io = io.BytesIO()

    def write_uint64(i):
        b_io.write(i.to_bytes(8, byteorder="little"))

    def write_int32(i):
        b_io.write(i.to_bytes(4, byteorder="little", signed=True))

    def write_uint8(i):
        b_io.write(i.to_bytes(1, byteorder="little"))

    b_io.write("PMTiles".encode())
    b_io.write(b"\x03")
    write_uint64(h["root_offset"])
    write_uint64(h["root_length"])
    write_uint64(h["metadata_offset"])
    write_uint64(h["metadata_length"])
    write_uint64(h.get("leaf_directory_offset", 0))
    write_uint64(h.get("leaf_directory_length", 0))
    write_uint64(h["tile_data_offset"])
    write_uint64(h["tile_data_length"])
    write_uint64(h.get("addressed_tiles_count", 0))
    write_uint64(h.get("tile_entries_count", 0))
    write_uint64(h.get("tile_contents_count", 0))
    b_io.write(b"\x01" if h["clustered"] else b"\x00")
    write_uint8(h["internal_compression"].value)
    write_uint8(h["tile_compression"].value)
    write_uint8(h["tile_type"].value)
    write_uint8(h["min_zoom"])
    write_uint8(h["max_zoom"])
    min_lon_e7 = h.get("min_lon_e7", int(-180 * 10000000))
    write_int32(min_lon_e7)
    min_lat_e7 = h.get("min_lat_e7", int(-90 * 10000000))
    write_int32(min_lat_e7)
    max_lon_e7 = h.get("max_lon_e7", int(180 * 10000000))
    write_int32(max_lon_e7)
    max_lat_e7 = h.get("max_lat_e7", int(90 * 10000000))
    write_int32(max_lat_e7)
    write_uint8(h.get("center_zoom",h["min_zoom"]))
    write_int32(h.get("center_lon_e7", round((min_lon_e7 + max_lon_e7) / 2)))
    write_int32(h.get("center_lat_e7", round((min_lat_e7 + max_lat_e7) / 2)))

    return b_io.getvalue()



import json
import tempfile
import gzip
import shutil
from contextlib import contextmanager
from .tile import (
    Entry,
    serialize_directory,
    Compression,
    serialize_header,
    tileid_to_zxy,
)


@contextmanager
def write(fname):
    f = open(fname, "wb")
    w = Writer(f)
    try:
        yield w
    finally:
        f.close()


def build_roots_leaves(entries, leaf_size):
    root_entries = []
    leaves_bytes = b""
    num_leaves = 0

    i = 0
    while i < len(entries):
        num_leaves += 1
        serialized = serialize_directory(entries[i : i + leaf_size])
        root_entries.append(
            Entry(entries[i].tile_id, len(leaves_bytes), len(serialized), 0)
        )
        leaves_bytes += serialized
        i += leaf_size

    return serialize_directory(root_entries), leaves_bytes, num_leaves


def optimize_directories(entries, target_root_len):
    test_bytes = serialize_directory(entries)
    if len(test_bytes) < target_root_len:
        return test_bytes, b"", 0

    leaf_size = 4096
    while True:
        root_bytes, leaves_bytes, num_leaves = build_roots_leaves(entries, leaf_size)
        if len(root_bytes) < target_root_len:
            return root_bytes, leaves_bytes, num_leaves
        leaf_size *= 2


class Writer:
    def __init__(self, f):
        self.f = f
        self.tile_entries = []
        self.hash_to_offset = {}
        self.tile_f = tempfile.TemporaryFile()
        self.offset = 0
        self.addressed_tiles = 0
        self.clustered = True

    def write_tile(self, tileid, data):
        if len(self.tile_entries) > 0 and tileid < self.tile_entries[-1].tile_id:
            self.clustered = False

        hsh = hash(data)
        if hsh in self.hash_to_offset:
            last = self.tile_entries[-1]
            found = self.hash_to_offset[hsh]
            if tileid == last.tile_id + last.run_length and last.offset == found:
                self.tile_entries[-1].run_length += 1
            else:
                self.tile_entries.append(Entry(tileid, found, len(data), 1))
        else:
            self.tile_f.write(data)
            self.tile_entries.append(Entry(tileid, self.offset, len(data), 1))
            self.hash_to_offset[hsh] = self.offset
            self.offset += len(data)

        self.addressed_tiles += 1

    def finalize(self, header, metadata):
        header["addressed_tiles_count"] = self.addressed_tiles
        header["tile_entries_count"] = len(self.tile_entries)
        header["tile_contents_count"] = len(self.hash_to_offset)

        self.tile_entries = sorted(self.tile_entries, key=lambda e: e.tile_id)

        header["min_zoom"] = tileid_to_zxy(self.tile_entries[0].tile_id)[0]
        header["max_zoom"] = tileid_to_zxy(self.tile_entries[-1].tile_id)[0]

        root_bytes, leaves_bytes, num_leaves = optimize_directories(
            self.tile_entries, 16384 - 127
        )

        compressed_metadata = gzip.compress(json.dumps(metadata).encode())
        header["clustered"] = self.clustered
        header["internal_compression"] = Compression.GZIP
        header["root_offset"] = 127
        header["root_length"] = len(root_bytes)
        header["metadata_offset"] = header["root_offset"] + header["root_length"]
        header["metadata_length"] = len(compressed_metadata)
        header["leaf_directory_offset"] = (
            header["metadata_offset"] + header["metadata_length"]
        )
        header["leaf_directory_length"] = len(leaves_bytes)
        header["tile_data_offset"] = (
            header["leaf_directory_offset"] + header["leaf_directory_length"]
        )
        header["tile_data_length"] = self.offset

        header_bytes = serialize_header(header)

        self.f.write(header_bytes)
        self.f.write(root_bytes)
        self.f.write(compressed_metadata)
        self.f.write(leaves_bytes)
        self.tile_f.seek(0)
        shutil.copyfileobj(self.tile_f, self.f)
        self.tile_f.close()



import argparse
import gzip
from vtiles.utils.pmtiles.writer import write
from vtiles.utils.pmtiles.tile import TileType, zxy_to_tileid, tileid_to_zxy, Compression
import sqlite3
from tqdm import tqdm

def mbtiles_to_header_json(mbtiles_metadata):
    header = {}

    header["min_zoom"] = int(mbtiles_metadata["minzoom"])

    header["max_zoom"] = int(mbtiles_metadata["maxzoom"])

    bounds = mbtiles_metadata["bounds"].split(",")
    header["min_lon_e7"] = int(float(bounds[0]) * 10000000)
    header["min_lat_e7"] = int(float(bounds[1]) * 10000000)
    header["max_lon_e7"] = int(float(bounds[2]) * 10000000)
    header["max_lat_e7"] = int(float(bounds[3]) * 10000000)

    center = mbtiles_metadata["center"].split(",")
    header["center_lon_e7"] = int(float(center[0]) * 10000000)
    header["center_lat_e7"] = int(float(center[1]) * 10000000)
    header["center_zoom"] = int(center[2])

    tile_format = mbtiles_metadata["format"]
    if tile_format == "pbf":
        header["tile_type"] = TileType.MVT
    elif tile_format == "png":
        header["tile_type"] = TileType.PNG
    elif tile_format == "jpeg":
        header["tile_type"] = TileType.JPEG
    elif tile_format == "webp":
        header["tile_type"] = TileType.WEBP
    elif tile_format == "avif":
        header["tile_type"] = TileType.AVIF
    else:
        header["tile_type"] = TileType.UNKNOWN

    if tile_format == "pbf" or mbtiles_metadata.get("compression") == "gzip":
        header["tile_compression"] = Compression.GZIP
    else:
        header["tile_compression"] = Compression.NONE

    return header, mbtiles_metadata

def mbtiles_to_pmtiles(input, output, maxzoom=None):
    conn = sqlite3.connect(input)
    cursor = conn.cursor()

    with write(output) as writer:
        # collect a set of all tile IDs
        tileid_set = []
        for row in cursor.execute(
            "SELECT zoom_level,tile_column,tile_row FROM tiles WHERE zoom_level <= ?",
            (maxzoom or 99,),
        ):
            flipped = (1 << row[0]) - 1 - row[2]
            tileid_set.append(zxy_to_tileid(row[0], row[1], flipped))

        tileid_set.sort()

        mbtiles_metadata = {}
        for row in cursor.execute("SELECT name,value FROM metadata"):
            mbtiles_metadata[row[0]] = row[1]
        is_pbf = mbtiles_metadata["format"] == "pbf"

        # query the db in ascending tile order
        for tileid in tqdm(tileid_set, desc="Converting tiles"):
            z, x, y = tileid_to_zxy(tileid)
            flipped = (1 << z) - 1 - y
            res = cursor.execute(
                "SELECT tile_data FROM tiles WHERE zoom_level = ? AND tile_column = ? AND tile_row = ?",
                (z, x, flipped),
            )
            data = res.fetchone()[0]
            # force gzip compression only for vector
            if is_pbf and data[0:2] != b"\x1f\x8b":
                data = gzip.compress(data)
            writer.write_tile(tileid, data)

        pmtiles_header, pmtiles_metadata = mbtiles_to_header_json(mbtiles_metadata)
        if maxzoom:
            pmtiles_header["max_zoom"] = int(maxzoom)
            mbtiles_metadata["maxzoom"] = maxzoom
        result = writer.finalize(pmtiles_header, pmtiles_metadata)

    conn.close()

def main():
    parser = argparse.ArgumentParser(description="Convert MBTiles to PMTiles.")
    parser.add_argument("-i", "--input", required=True, help="Input MBTiles file path.")
    parser.add_argument("-o", "--output", required=True, help="Output PMTiles file path.")
    parser.add_argument("-z", "--maxzoom", type=int, help="Maximum zoom level to process.")
    args = parser.parse_args()

    mbtiles_to_pmtiles(args.input, args.output, args.maxzoom)

if __name__ == "__main__":
    main()



import sqlite3
import argparse
import gzip, zlib

def extract_tile_to_pbf(mbtiles_file, z, x, y, output_pbf):
    try:
        # Connect to the MBTiles database
        conn = sqlite3.connect(mbtiles_file)
        cursor = conn.cursor()        
       
        # Query the tile data
        cursor.execute('SELECT tile_data FROM tiles WHERE zoom_level=? AND tile_column=? AND tile_row=?', (z, x, y))
        row = cursor.fetchone()
        conn.close()
        
        if row:
            tile_data = row[0] 
            # Write the decompressed tile data to the output PBF file
            with open(output_pbf, 'wb') as f:
                f.write(tile_data)
            print(f"Tile {z}/{x}/{y} successfully extracted to {output_pbf}")
        else:
            print("Tile not found in MBTiles file")
    except sqlite3.Error as e:
        print(f"Failed to read MBTiles file: {e}")

def main():
    parser = argparse.ArgumentParser(description='Extract a tile from MBTiles to PBF.')
    parser.add_argument('-i', '--input', required=True, help='Input MBTiles file')
    parser.add_argument('-z', '--zoom', type=int, required=True, help='Zoom level')
    parser.add_argument('-x', '--x', type=int, required=True, help='tile column')
    parser.add_argument('-y', '--y', type=int, required=True, help='tile row')
    parser.add_argument('-o', '--output', required=True, help='Output PBF file')
    
    args = parser.parse_args()
    
    extract_tile_to_pbf(args.input, args.zoom, args.x, args.y, args.output)

if __name__ == '__main__':
    main()



import osmium as o
import sys
import os
import datetime

class FileStatsHandler(o.SimpleHandler):
    def __init__(self):
        super(FileStatsHandler, self).__init__()
        self.nodes = 0
        self.ways = 0
        self.rels = 0

    def node(self, n):
        self.nodes += 1

    def way(self, w):
        self.ways += 1

    def relation(self, r):
        self.rels += 1

class CountHandler(o.SimpleHandler):
    def __init__(self):
        super(CountHandler, self).__init__()
        self.nodes = 0
        self.ways = 0
        self.rels = 0

    def node(self, n):
        self.nodes += 1

    def way(self, w):
        self.ways += 1

    def relation(self, r):
        self.rels += 1

def main():
    if len(sys.argv) != 2:
        print(f"Usage: {sys.argv[0]} <path_to_osm_file.osm.pbf>")
        sys.exit(1)

    osmfile = sys.argv[1]

    if not os.path.exists(osmfile):
        print(f"File not found: {osmfile}")
        sys.exit(1)

    file_stat = os.stat(osmfile)
    file_size_mb = round(file_stat.st_size / (1024 * 1024), 2)
    file_last_modified = datetime.datetime.fromtimestamp(file_stat.st_mtime).strftime("%Y-%m-%d %I:%M:%S %p")

    print('######')
    print(f"File size: {file_size_mb} MB")
    print(f"Last modified: {file_last_modified}")
    print('######')

    count_handler = CountHandler()
    try:
        # First pass to count total nodes, ways, and relations
        count_handler.apply_file(osmfile)

        # Initialize the FileStatsHandler without progress bars
        handler = FileStatsHandler()
        handler.apply_file(osmfile)

    except Exception as e:
        print(f"Error while processing the file: {e}")
        sys.exit(1)
    
    finally:
        sys.stdout.flush()  # Explicitly flush stdout again
        sys.exit(0)

if __name__ == "__main__":
    main()



import http.server
import socketserver
import re
import psycopg2
import yaml
import argparse

# Argument parser for command-line arguments
parser = argparse.ArgumentParser(description='Serve MVT tiles from a PostgreSQL/PostGIS database.')
parser.add_argument('-config', help='Path to the YAML configuration file')
args = parser.parse_args()

# Load configuration from a YAML file
with open(args.config_file, 'r') as file:
    config = yaml.safe_load(file)

# Extract database and HTTP server configuration
DATABASE = config['database']
HTTP_SERVER = config['http_server']

########################################################################

class TileRequestHandler(http.server.BaseHTTPRequestHandler):

    DATABASE_CONNECTION = None

    def pathToTile(self, path):
        m = re.search(r'^\/(\d+)\/(\d+)\/(\d+)\.(\w+)', path)
        if m:
            return {'zoom': int(m.group(1)),
                    'x': int(m.group(2)),
                    'y': int(m.group(3)),
                    'format': m.group(4)}
        else:
            return None

    def tileIsValid(self, tile):
        if not ('x' in tile and 'y' in tile and 'zoom' in tile):
            return False
        if 'format' not in tile or tile['format'] not in ['pbf', 'mvt']:
            return False
        size = 2 ** tile['zoom']
        if tile['x'] >= size or tile['y'] >= size:
            return False
        if tile['x'] < 0 or tile['y'] < 0:
            return False
        return True

    def envelopeToBoundsSQL(self, tile):
        return f'ST_TileEnvelope({tile["zoom"]}, {tile["x"]}, {tile["y"]})'

    # def envelopeToSQL(self, tile):
    #     sql_queries = []
    #     for table_name, table_config in config['tables'].items():
    #         if 'min_zoom' in table_config and tile['zoom'] < table_config['min_zoom']:
    #             continue
    #         if 'max_zoom' in table_config and tile['zoom'] > table_config['max_zoom']:
    #             continue
            
    #         tbl = table_config.copy()
    #         tbl['env'] = self.envelopeToBoundsSQL(tile)
    #         sql_tmpl = """
    #             WITH 
    #             bounds AS (
    #                 SELECT {env} AS geom, 
    #                     {env}::box2d AS b2d
    #             ),
    #             mvtgeom AS (
    #                 SELECT ST_AsMVTGeom(ST_Transform(t.{geomColumn}, 3857), bounds.b2d) AS geom, 
    #                     {attrColumns}
    #                 FROM {table} t, bounds
    #                 WHERE ST_Intersects(t.{geomColumn}, ST_Transform(bounds.geom, {srid}))
    #             ) 
    #             SELECT '{table_name}' AS layer, ST_AsMVT(mvtgeom.*, '{table_name}') FROM mvtgeom
    #         """
    #         sql_queries.append(sql_tmpl.format(table_name=table_name, **tbl))
        
    #     return sql_queries

    def envelopeToSQL(self, tile):
        sql_queries = []
        for table_name, table_config in config['tables'].items():
            if 'min_zoom' in table_config and tile['zoom'] < table_config['min_zoom']:
                continue
            if 'max_zoom' in table_config and tile['zoom'] > table_config['max_zoom']:
                continue
            
            tbl = table_config.copy()
            area_condition = ""
            if 'area_thresholds' in tbl:
                for threshold in tbl['area_thresholds']:
                    if tile['zoom'] == threshold['zoom']:
                        area_condition = f" AND ST_Area(t.{tbl['geomColumn']}) > {threshold['area']}"
                        break

            sql_tmpl = f"""
                WITH 
                bounds AS (
                    SELECT ST_TileEnvelope({tile["zoom"]}, {tile["x"]}, {tile["y"]}) AS geom, 
                        ST_TileEnvelope({tile["zoom"]}, {tile["x"]}, {tile["y"]})::box2d AS b2d
                ),
                mvtgeom AS (
                    SELECT ST_AsMVTGeom(ST_Transform(t.{tbl["geomColumn"]}, 3857), bounds.b2d) AS geom, 
                        {tbl["attrColumns"]}
                    FROM {tbl["table"]} t, bounds
                    WHERE ST_Intersects(t.{tbl["geomColumn"]}, ST_Transform(bounds.geom, {tbl["srid"]}))
                    {area_condition}
                ) 
                SELECT '{table_name}' AS layer, ST_AsMVT(mvtgeom.*, '{table_name}') FROM mvtgeom
            """
            sql_queries.append(sql_tmpl)
        
        return sql_queries

    def sqlToPbf(self, sql_queries):
        if not self.DATABASE_CONNECTION:
            try:
                self.DATABASE_CONNECTION = psycopg2.connect(**DATABASE)
            except (Exception, psycopg2.Error) as error:
                self.send_error(500, f"Cannot connect to database: {str(error)}")
                return None

        responses = []
        with self.DATABASE_CONNECTION.cursor() as cur:
            for sql in sql_queries:
                cur.execute(sql)
                if not cur:
                    self.send_error(404, f"SQL query failed: {sql}")
                    return None
                responses.append(cur.fetchone())

        return responses

    def do_GET(self):
        tile = self.pathToTile(self.path)
        if not (tile and self.tileIsValid(tile)):
            self.send_error(400, f"Invalid tile path: {self.path}")
            return

        sql_queries = self.envelopeToSQL(tile)
        pbf_responses = self.sqlToPbf(sql_queries)

        self.log_message("path: %s\ntile: %s" % (self.path, tile))
        self.log_message("sql_queries: %s" % (sql_queries))

        if pbf_responses is None or len(pbf_responses) == 0:
            self.send_error(500, "Failed to generate tile data")
            return
        
        self.send_response(200)
        self.send_header("Access-Control-Allow-Origin", "*")
        self.send_header("Content-type", "application/x-protobuf")
        self.end_headers()
        
        # Concatenate all MVT responses into a single protobuf response
        for response in pbf_responses:
            self.wfile.write(response[1])  # Assuming the MVT data is in the second column

########################################################################

def main():
    try:
        with http.server.HTTPServer((HTTP_SERVER['host'], HTTP_SERVER['port']), TileRequestHandler) as server:
            print(f"Serving at port {HTTP_SERVER['port']}")
            server.serve_forever()
    except KeyboardInterrupt:
        if TileRequestHandler.DATABASE_CONNECTION:
            TileRequestHandler.DATABASE_CONNECTION.close()
        print('^C received, shutting down server')
        server.socket.close()

if __name__ == "__main__":
    main()



import os
import sys
import boto3
from tqdm import tqdm
from concurrent.futures import ThreadPoolExecutor
import multiprocessing
import argparse
import logging

# logging.basicConfig(level=logging.INFO)
logging.basicConfig(level=logging.ERROR)
s3 = boto3.client('s3')

def upload_file(bucket_name, local_file_path, s3_key, content_type=None, content_encoding=None):
    try:
        extra_args = {}
        if content_type:
            extra_args['ContentType'] = content_type
        if content_encoding:
            extra_args['ContentEncoding'] = content_encoding
        s3.upload_file(local_file_path, bucket_name, s3_key, ExtraArgs=extra_args)
        return True
    except Exception as e:
        # logging.error(f"Error uploading {local_file_path} to {s3_key}: {e}")
        print(f"Error uploading {local_file_path} to {s3_key}: {e}")
        return False

def upload_files(bucket_name, input_folder, s3_prefix='', content_type=None, content_encoding=None):
    total_files = sum(len(files) for _, _, files in os.walk(input_folder))
    num_cores = multiprocessing.cpu_count()

    with tqdm(total=total_files, desc="Uploading", unit="file") as pbar:
        with ThreadPoolExecutor(max_workers=num_cores*2) as executor:
            futures = []
            for root, _, files in os.walk(input_folder):
                for file in files:
                    local_file_path = os.path.join(root, file)
                    s3_key = os.path.relpath(local_file_path, input_folder)
                    s3_key = os.path.join(s3_prefix, s3_key).replace('\\', '/')  # for Windows compatibility
                    future = executor.submit(upload_file, bucket_name, local_file_path, s3_key, content_type, content_encoding)
                    future.add_done_callback(lambda p: pbar.update())
                    futures.append(future)
            # Wait for all uploads to complete
            for future in futures:
                future.result()

def folder2s3(input_folder, format='', bucket_name='', s3_prefix='', aws_access_key_id=None, aws_secret_access_key=None, aws_region=None):
    session = boto3.Session(
        region_name=aws_region,
        aws_access_key_id=aws_access_key_id,
        aws_secret_access_key=aws_secret_access_key,
    )
    try:
        logging.info('Creating a connection to S3')
        global s3
        s3 = session.client('s3')
    except Exception as e:
        logging.error(f"Failed to create S3 client: {e}")
        return

    try:
        logging.info(f'Uploading folder {input_folder} to S3 bucket: {bucket_name}. Press Ctrl+C to cancel')
        if format == 'pbf' or format == 'mvt' :
            upload_files(bucket_name, input_folder, s3_prefix, 'application/x-protobuf', 'gzip')
        else:
            upload_files(bucket_name, input_folder, s3_prefix)
        logging.info('Uploading folder to S3 done!')
    except Exception as e:
        logging.error(f"Error uploading folder to S3: {e}")
        return

def main():
    parser = argparse.ArgumentParser(description='Upload a tiles folder to S3.')
    parser.add_argument('-i', type=str, required=True, help='The tiles folder to upload.')
    parser.add_argument('-format', type=str, required=True, choices=['pbf', 'png', 'jpg', 'jpeg', 'webp', 'pbf', 'mvt'], help='format of the files to upload.')
    args = parser.parse_args()

    input_folder = args.i
    format = args.format

    if not os.path.exists(input_folder) or not os.path.isdir(input_folder):
        logging.error('Input folder does not exist or is invalid. Please recheck and provide a correct one.')
        exit()

    input_folder_abspath = os.path.abspath(input_folder)

    logging.info('### Input S3 parameters:')
    s3_bucket_name = input('S3 Bucket name: ')
    while not s3_bucket_name:
        logging.error('S3 Bucket name is required.')
        s3_bucket_name = input('S3 Bucket name: ')

    s3_prefix = input(f'S3 prefix (Press Enter to upload to the bucket {s3_bucket_name} root folder): ')
    if not s3_prefix:
        s3_prefix = ''

    aws_access_key_id = input('AWS Access Key ID: ')
    while not aws_access_key_id:
        aws_access_key_id = input('AWS Access Key ID is required. Please input AWS Access Key ID: ')

    aws_secret_access_key = input('AWS Secret Access Key: ')
    while not aws_secret_access_key:
        aws_secret_access_key = input('AWS Secret Access Key is required. Please input AWS Secret Access Key: ')

    aws_region = input('AWS region (Press Enter to choose default region): ')
    if not aws_region:
        aws_region = None

    folder2s3(input_folder_abspath, format, s3_bucket_name, s3_prefix, aws_access_key_id, aws_secret_access_key, aws_region)

if __name__ == "__main__":
    main()



import os
import shutil
import sqlite3
import argparse
import logging

# Set up logging
logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(levelname)s - %(message)s')

def remove_duplicates(input_mbtiles, output_mbtiles):
    # Check if the output file exists and remove it
    if os.path.exists(output_mbtiles):
        os.remove(output_mbtiles)
        logging.info(f"Removed existing output file: {output_mbtiles}")

    # Copy the original MBTiles file to the output path
    shutil.copyfile(input_mbtiles, output_mbtiles)
    logging.info(f"Copied input file to output: {input_mbtiles} -> {output_mbtiles}")

    conn = sqlite3.connect(output_mbtiles)
    cursor = conn.cursor()

    try:
        # Check if 'tiles' is a view
        cursor.execute("SELECT type FROM sqlite_master WHERE name = 'tiles'")
        result = cursor.fetchone()

        if result and result[0] == 'view':
            logging.info("'tiles' is a view. Converting to a table...")

            # Step 1: Create a temporary table from the view
            cursor.execute("CREATE TABLE IF NOT EXISTS temp_tiles AS SELECT * FROM tiles")
            cursor.execute("DROP VIEW IF EXISTS tiles")
        else:
            logging.info("'tiles' is a table. Proceeding with the table directly.")
            cursor.execute("ALTER TABLE tiles RENAME TO temp_tiles")

        # Step 2: Remove duplicates by keeping only the first occurrence
        logging.info("Removing duplicates from the 'temp_tiles' table...")

        # Create a new table to store unique rows
        cursor.execute("""
            CREATE TABLE IF NOT EXISTS clean_tiles AS 
            SELECT zoom_level, tile_column, tile_row, tile_data 
            FROM temp_tiles
            GROUP BY zoom_level, tile_column, tile_row
        """)

        # Drop the temporary table with duplicates
        cursor.execute("DROP TABLE IF EXISTS temp_tiles")

        # Rename the clean table to 'tiles'
        cursor.execute("ALTER TABLE clean_tiles RENAME TO tiles")

        # Add a unique index to prevent future duplicates
        cursor.execute("""
            CREATE UNIQUE INDEX tile_index
            ON tiles(zoom_level, tile_column, tile_row)
        """)

        # Commit the changes
        conn.commit()
        logging.info("Duplicates removed and 'tiles' table updated successfully.")

    except sqlite3.Error as e:
        logging.error(f"An error occurred: {e}")
    finally:
        # Close the connection
        conn.close()

def main():
    # Create argument parser inside the main function
    parser = argparse.ArgumentParser(description="Remove duplicate tiles from an MBTiles file.")
    parser.add_argument('-i', '--input', required=True, help="Input MBTiles file")
    parser.add_argument('-o', '--output', required=True, help="Output MBTiles file")

    # Parse the command-line arguments
    args = parser.parse_args()

    # Show progress for copying the file
    logging.info("Starting to remove duplicates...")
    remove_duplicates(args.input, args.output)
    logging.info("Processing complete.")

if __name__ == '__main__':
    main()



# BASE
URL = "http://download.geofabrik.de/"
suffix = "-latest.osm.pbf"
africa_url = "africa/"
asia_url = "asia/"
australia_oceania_url = "australia-oceania/"
central_america_url = "central-america/"
europe_url = "europe/"
north_america_url = "north-america/"
south_america_url = "south-america/"

baden_wuerttemberg_url = "europe/germany/baden-wuerttemberg/"
bayern_url = "europe/germany/bayern/"
brazil_url = "south-america/brazil/"
canada_url = "north-america/canada/"
england_url = "europe/great-britain/england/"
france_url = "europe/france/"
gb_url = "europe/great-britain/"
germany_url = "europe/germany/"
italy_url = "europe/italy/"
japan_url = "asia/japan/"
netherlands_url = "europe/netherlands/"
nordrhein_wesfalen_url = "europe/germany/nordrhein-westfalen/"
poland_url = "europe/poland/"
russia_url = "russia/"
usa_url = "north-america/us/"


class USA:
    # State level data sources
    regions = [
        "alabama",
        "alaska",
        "arizona",
        "arkansas",
        "colorado",
        "connecticut",
        "delaware",
        "district_of_columbia",
        "florida",
        "georgia",
        "hawaii",
        "idaho",
        "illinois",
        "indiana",
        "iowa",
        "kansas",
        "kentucky",
        "louisiana",
        "maine",
        "maryland",
        "massachusetts",
        "michigan",
        "minnesota",
        "mississippi",
        "missouri",
        "montana",
        "nebraska",
        "nevada",
        "new_hampshire",
        "new_mexico",
        "new_york",
        "new_jersey",
        "north_carolina",
        "north_dakota",
        "ohio",
        "oklahoma",
        "oregon",
        "pennsylvania",
        "puerto_rico",
        "rhode_island",
        "south_carolina",
        "south_dakota",
        "tennessee",
        "texas",
        "utah",
        "vermont",
        "virginia",
        "washington",
        "west_virginia",
        "wisconsin",
        "wyoming",
    ]

    available = regions + ["southern_california", "northern_california"]
    available.sort()

    country = {"name": "us" + suffix, "url": URL + north_america_url + "us" + suffix}

    # Create data sources
    _sources = {
        region: {
            "name": region.replace("_", "-") + suffix,
            "url": URL + usa_url + region.replace("_", "-") + suffix,
        }
        for region in regions
    }

    # Add California separately
    _sources["southern_california"] = {
        "name": "socal" + suffix,
        "url": URL + "north-america/us/california/socal-latest.osm.pbf",
    }
    _sources["northern_california"] = {
        "name": "norcal" + suffix,
        "url": URL + "north-america/us/california/norcal-latest.osm.pbf",
    }
    __dict__ = _sources

    def __getattr__(self, name):
        return self.__dict__[name]

    def __call__(self):
        return self.available


class France:
    regions = [
        "alsace",
        "aquitaine",
        "auvergne",
        "basse_normandie",
        "bourgogne",
        "bretagne",
        "centre",
        "champagne_ardenne",
        "corse",
        "franche_comte",
        "guadeloupe",
        "guyane",
        "haute_normandie",
        "ile_de_france",
        "languedoc_roussillon",
        "limousin",
        "lorraine",
        "martinique",
        "mayotte",
        "midi_pyrenees",
        "nord_pas_de_calais",
        "pays_de_la_loire",
        "picardie",
        "poitou_charentes",
        "provence_alpes_cote_d_azur",
        "reunion",
        "rhone_alpes",
    ]

    available = regions
    available.sort()

    country = {"name": "france" + suffix, "url": URL + europe_url + "france" + suffix}

    # Create data sources
    _sources = {
        region: {
            "name": region.replace("_", "-") + suffix,
            "url": URL + france_url + region.replace("_", "-") + suffix,
        }
        for region in regions
    }

    __dict__ = _sources

    def __getattr__(self, name):
        return self.__dict__[name]

    def __call__(self):
        return self.available


class England:
    regions = [
        "bedfordshire",
        "berkshire",
        "bristol",
        "buckinghamshire",
        "cambridgeshire",
        "cheshire",
        "cornwall",
        "cumbria",
        "derbyshire",
        "devon",
        "dorset",
        "durham",
        "east_sussex",
        "east_yorkshire_with_hull",
        "essex",
        "gloucestershire",
        "greater_london",
        "greater_manchester",
        "hampshire",
        "herefordshire",
        "hertfordshire",
        "isle_of_wight",
        "kent",
        "lancashire",
        "leicestershire",
        "lincolnshire",
        "merseyside",
        "norfolk",
        "north_yorkshire",
        "northamptonshire",
        "northumberland",
        "nottinghamshire",
        "oxfordshire",
        "rutland",
        "shropshire",
        "somerset",
        "south_yorkshire",
        "staffordshire",
        "suffolk",
        "surrey",
        "tyne_and_wear",
        "warwickshire",
        "west_midlands",
        "west_sussex",
        "west_yorkshire",
        "wiltshire",
        "worcestershire",
    ]

    available = regions
    available.sort()

    # Create data sources
    _sources = {
        region: {
            "name": region.replace("_", "-") + suffix,
            "url": URL + england_url + region.replace("_", "-") + suffix,
        }
        for region in regions
    }

    __dict__ = _sources

    def __getattr__(self, name):
        return self.__dict__[name]

    def __call__(self):
        return self.available


class GreatBritain:
    regions = ["england", "scotland", "wales"]
    england = England()

    available = regions + england.available
    available.sort()

    country = {
        "name": "great-britain" + suffix,
        "url": URL + europe_url + "great-britain" + suffix,
    }

    # Create data sources
    _sources = {
        region: {
            "name": region.replace("_", "-") + suffix,
            "url": URL + gb_url + region.replace("_", "-") + suffix,
        }
        for region in regions
    }

    for region in england.available:
        _sources[region] = {
            "name": region.replace("_", "-") + suffix,
            "url": URL + england_url + region.replace("_", "-") + suffix,
        }

    __dict__ = _sources

    def __getattr__(self, name):
        return self.__dict__[name]

    def __call__(self):
        return self.available


class Italy:
    regions = ["centro", "isole", "nord_est", "nord_ovest", "sud"]

    available = regions
    available.sort()

    country = {"name": "italy" + suffix, "url": URL + europe_url + "italy" + suffix}

    # Create data sources
    _sources = {
        region: {
            "name": region.replace("_", "-") + suffix,
            "url": URL + italy_url + region.replace("_", "-") + suffix,
        }
        for region in regions
    }

    __dict__ = _sources

    def __getattr__(self, name):
        return self.__dict__[name]

    def __call__(self):
        return self.available


class Russia:
    regions = [
        "central_fed_district",
        "crimean_fed_district",
        "far_eastern_fed_district",
        "kaliningrad",
        "north_caucasus_fed_district",
        "northwestern_fed_district",
        "siberian_fed_district",
        "south_fed_district",
        "ural_fed_district",
        "volga_fed_district",
    ]

    available = regions
    available.sort()

    country = {"name": "russia" + suffix, "url": URL + russia_url + "russia" + suffix}

    # Create data sources
    _sources = {
        region: {
            "name": region.replace("_", "-") + suffix,
            "url": URL + russia_url + region.replace("_", "-") + suffix,
        }
        for region in regions
    }

    __dict__ = _sources

    def __getattr__(self, name):
        return self.__dict__[name]

    def __call__(self):
        return self.available


class Poland:
    regions = [
        "dolnoslaskie",
        "kujawsko_pomorskie",
        "lodzkie",
        "lubelskie",
        "lubuskie",
        "malopolskie",
        "mazowieckie",
        "opolskie",
        "podkarpackie",
        "podlaskie",
        "pomorskie",
        "slaskie",
        "swietokrzyskie",
        "warminsko_mazurskie",
        "wielkopolskie",
        "zachodniopomorskie",
    ]

    available = regions
    available.sort()

    country = {"name": "poland" + suffix, "url": URL + europe_url + "poland" + suffix}

    # Create data sources
    _sources = {
        region: {
            "name": region.replace("_", "-") + suffix,
            "url": URL + poland_url + region.replace("_", "-") + suffix,
        }
        for region in regions
    }

    __dict__ = _sources

    def __getattr__(self, name):
        return self.__dict__[name]

    def __call__(self):
        return self.available


class BadenWuerttemberg:
    regions = [
        "freiburg_regbez",
        "karlsruhe_regbez",
        "stuttgart_regbez",
        "tuebingen_regbez",
    ]

    available = regions
    available.sort()

    # Create data sources
    _sources = {
        region: {
            "name": region.replace("_", "-") + suffix,
            "url": URL + baden_wuerttemberg_url + region.replace("_", "-") + suffix,
        }
        for region in regions
    }

    __dict__ = _sources

    def __getattr__(self, name):
        return self.__dict__[name]

    def __call__(self):
        return self.available


class NordrheinWestfalen:
    regions = [
        "arnsberg_regbez",
        "detmold_regbez",
        "duesseldorf_regbez",
        "koeln_regbez",
        "muenster_regbez",
    ]

    available = regions
    available.sort()

    # Create data sources
    _sources = {
        region: {
            "name": region.replace("_", "-") + suffix,
            "url": URL + nordrhein_wesfalen_url + region.replace("_", "-") + suffix,
        }
        for region in regions
    }

    __dict__ = _sources

    def __getattr__(self, name):
        return self.__dict__[name]

    def __call__(self):
        return self.available


class Bayern:
    regions = [
        "mittelfranken",
        "niederbayern",
        "oberbayern",
        "oberfranken",
        "oberpfalz",
        "schwaben",
        "unterfranken",
    ]

    available = regions
    available.sort()

    # Create data sources
    _sources = {
        region: {
            "name": region.replace("_", "-") + suffix,
            "url": URL + bayern_url + region.replace("_", "-") + suffix,
        }
        for region in regions
    }

    __dict__ = _sources

    def __getattr__(self, name):
        return self.__dict__[name]

    def __call__(self):
        return self.available


class Germany:
    regions = [
        "baden_wuerttemberg",
        "bayern",
        "berlin",
        "brandenburg",
        "bremen",
        "hamburg",
        "hessen",
        "mecklenburg_vorpommern",
        "niedersachsen",
        "nordrhein_westfalen",
        "rheinland_pfalz",
        "saarland",
        "sachsen_anhalt",
        "sachsen",
        "schleswig_holstein",
        "thueringen",
    ]

    baden_wuerttemberg = BadenWuerttemberg()
    bayern = Bayern()
    nordrhein_westfalen = NordrheinWestfalen()

    available = (
        regions
        + bayern.available
        + baden_wuerttemberg.available
        + nordrhein_westfalen.available
    )
    available.sort()

    country = {"name": "germany" + suffix, "url": URL + europe_url + "germany" + suffix}

    # Create data sources
    _sources = {
        region: {
            "name": region.replace("_", "-") + suffix,
            "url": URL + germany_url + region.replace("_", "-") + suffix,
        }
        for region in regions
    }

    for region in nordrhein_westfalen.available:
        _sources[region] = {
            "name": region.replace("_", "-") + suffix,
            "url": URL + nordrhein_wesfalen_url + region.replace("_", "-") + suffix,
        }

    for region in baden_wuerttemberg.available:
        _sources[region] = {
            "name": region.replace("_", "-") + suffix,
            "url": URL + baden_wuerttemberg_url + region.replace("_", "-") + suffix,
        }

    for region in bayern.available:
        _sources[region] = {
            "name": region.replace("_", "-") + suffix,
            "url": URL + bayern_url + region.replace("_", "-") + suffix,
        }

    __dict__ = _sources

    def __getattr__(self, name):
        return self.__dict__[name]

    def __call__(self):
        return self.available


class Netherlands:
    regions = [
        "drenthe",
        "flevoland",
        "friesland",
        "gelderland",
        "groningen",
        "limburg",
        "noord_brabant",
        "noord_holland",
        "overijssel",
        "utrecht",
        "zeeland",
        "zuid_holland",
    ]
    available = regions
    available.sort()

    country = {
        "name": "netherlands" + suffix,
        "url": URL + europe_url + "netherlands" + suffix,
    }

    # Create data sources
    _sources = {
        region: {
            "name": region.replace("_", "-") + suffix,
            "url": URL + netherlands_url + region.replace("_", "-") + suffix,
        }
        for region in regions
    }

    __dict__ = _sources

    def __getattr__(self, name):
        return self.__dict__[name]

    def __call__(self):
        return self.available


class Canada:
    regions = [
        "alberta",
        "british_columbia",
        "manitoba",
        "new_brunswick",
        "newfoundland_and_labrador",
        "northwest_territories",
        "nova_scotia",
        "nunavut",
        "ontario",
        "prince_edward_island",
        "quebec",
        "saskatchewan",
        "yukon",
    ]
    available = regions
    available.sort()

    country = {
        "name": "canada" + suffix,
        "url": URL + north_america_url + "canada" + suffix,
    }

    # Create data sources
    _sources = {
        region: {
            "name": region.replace("_", "-") + suffix,
            "url": URL + canada_url + region.replace("_", "-") + suffix,
        }
        for region in regions
    }

    __dict__ = _sources

    def __getattr__(self, name):
        return self.__dict__[name]

    def __call__(self):
        return self.available


class Brazil:
    regions = ["centro_oeste", "nordeste", "norte", "sudeste", "sul"]
    available = regions
    available.sort()

    country = {
        "name": "brazil" + suffix,
        "url": URL + south_america_url + "brazil" + suffix,
    }

    # Create data sources
    _sources = {
        region: {
            "name": region.replace("_", "-") + suffix,
            "url": URL + brazil_url + region.replace("_", "-") + suffix,
        }
        for region in regions
    }

    __dict__ = _sources

    def __getattr__(self, name):
        return self.__dict__[name]

    def __call__(self):
        return self.available


class Japan:
    regions = [
        "chubu",
        "chugoku",
        "hokkaido",
        "kansai",
        "kanto",
        "kyushu",
        "shikoku",
        "tohoku",
    ]

    available = regions
    available.sort()

    country = {"name": "japan" + suffix, "url": URL + asia_url + "japan" + suffix}

    # Create data sources
    _sources = {
        region: {
            "name": region.replace("_", "-") + suffix,
            "url": URL + japan_url + region + suffix,
        }
        for region in regions
    }

    __dict__ = _sources

    def __getattr__(self, name):
        return self.__dict__[name]

    def __call__(self):
        return self.available


class AustraliaOceania:
    regions = [
        "australia",
        "cook_islands",
        "fiji",
        "ile-de-clipperton",
        "kiribati",
        "marshall_islands",
        "micronesia",
        "nauru",
        "new_caledonia",
        "new_zealand",
        "niue",
        "palau",
        "papua_new_guinea",
        "pitcairn-islands",
        "polynesie-francaise",
        "samoa",
        "solomon_islands",
        "tokelau",
        "tonga",
        "tuvalu",
        "vanuatu",
        "wallis-et-futuna",
    ]

    available = regions
    available.sort()

    continent = {
        "name": "australia-oceania" + suffix,
        "url": URL + "australia-oceania" + suffix,
    }

    # Create data sources
    _sources = {
        region: {
            "name": region.replace("_", "-") + suffix,
            "url": URL + australia_oceania_url + region.replace("_", "-") + suffix,
        }
        for region in regions
    }

    __dict__ = _sources

    def __getattr__(self, name):
        return self.__dict__[name]

    def __call__(self):
        return self.available


class NorthAmerica:
    regions = [
        "canada",
        "greenland",
        "mexico",
        "usa",
        "us_midwest",
        "us_northeast",
        "us_pacific",
        "us_south",
        "us_west",
    ]

    usa = USA()
    canada = Canada()

    available = regions
    available.sort()

    continent = {
        "name": "north-america" + suffix,
        "url": URL + "north-america" + suffix,
    }

    # Create data sources
    _sources = {
        region: {
            "name": region.replace("_", "-") + suffix,
            "url": URL + north_america_url + region.replace("_", "-") + suffix,
        }
        for region in regions
        if region != "usa"
    }
    # USA is "us" in GeoFabrik
    _sources["usa"] = {
        "name": "us" + suffix,
        "url": URL + north_america_url + "us" + suffix,
    }

    __dict__ = _sources

    def __getattr__(self, name):
        return self.__dict__[name]

    def __call__(self):
        return self.available


class SouthAmerica:
    regions = [
        "argentina",
        "bolivia",
        "brazil",
        "chile",
        "colombia",
        "ecuador",
        "paraguay",
        "peru",
        "suriname",
        "uruguay",
        "venezuela",
    ]

    brazil = Brazil()

    available = regions
    available.sort()
    available.sort()

    continent = {
        "name": "south-america" + suffix,
        "url": URL + "south-america" + suffix,
    }

    # Create data sources
    _sources = {
        region: {
            "name": region.replace("_", "-") + suffix,
            "url": URL + south_america_url + region + suffix,
        }
        for region in regions
    }

    __dict__ = _sources

    def __getattr__(self, name):
        return self.__dict__[name]

    def __call__(self):
        return self.available


class CentralAmerica:
    regions = [
        "bahamas",
        "belize",
        "costa-rica",
        "cuba",
        "el-salvador",
        "guatemala",
        "haiti_and_domrep",
        "honduras",
        "jamaica",
        "nicaragua",
        "panama",
    ]

    available = regions
    available.sort()

    continent = {
        "name": "central-america" + suffix,
        "url": URL + "central-america" + suffix,
    }

    # Create data sources
    _sources = {
        region: {
            "name": region.replace("_", "-") + suffix,
            "url": URL + central_america_url + region.replace("_", "-") + suffix,
        }
        for region in regions
    }

    __dict__ = _sources

    def __getattr__(self, name):
        return self.__dict__[name]

    def __call__(self):
        return self.available


class Europe:
    # Country specific subregions
    france = France()
    great_britain = GreatBritain()
    italy = Italy()
    russia = Russia()
    poland = Poland()
    germany = Germany()
    netherlands = Netherlands()

    regions = [
        "albania",
        "andorra",
        "austria",
        "azores",
        "belarus",
        "belgium",
        "bosnia_herzegovina",
        "bulgaria",
        "croatia",
        "cyprus",
        "czech_republic",
        "denmark",
        "estonia",
        "faroe_islands",
        "finland",
        "france",
        "georgia",
        "germany",
        "great_britain",
        "greece",
        "hungary",
        "iceland",
        "ireland_and_northern_ireland",
        "isle_of_man",
        "italy",
        "kosovo",
        "latvia",
        "liechtenstein",
        "lithuania",
        "luxembourg",
        "macedonia",
        "malta",
        "moldova",
        "monaco",
        "montenegro",
        "netherlands",
        "norway",
        "poland",
        "portugal",
        "romania",
        "russia",
        "serbia",
        "slovakia",
        "slovenia",
        "spain",
        "sweden",
        "switzerland",
        "turkey",
        "ukraine",
    ]

    available = regions
    available.sort()

    continent = {"name": "europe" + suffix, "url": URL + "europe" + suffix}

    # Create data sources
    _sources = {
        region: {
            "name": region.replace("_", "-") + suffix,
            "url": URL + europe_url + region.replace("_", "-") + suffix,
        }
        for region in regions
        if region != "russia"
    }
    # Russia is separately from Europe
    _sources["russia"] = {"name": "russia" + suffix, "url": URL + "russia" + suffix}

    __dict__ = _sources

    def __getattr__(self, name):
        return self.__dict__[name]

    def __call__(self):
        return self.available


class Africa:
    regions = [
        "algeria",
        "angola",
        "benin",
        "botswana",
        "burkina_faso",
        "burundi",
        "cameroon",
        "canary_islands",
        "cape_verde",
        "central_african_republic",
        "chad",
        "comores",
        "congo_brazzaville",
        "congo_democratic_republic",
        "djibouti",
        "egypt",
        "equatorial_guinea",
        "eritrea",
        "ethiopia",
        "gabon",
        "ghana",
        "guinea_bissau",
        "guinea",
        "ivory_coast",
        "kenya",
        "lesotho",
        "liberia",
        "libya",
        "madagascar",
        "malawi",
        "mali",
        "mauritania",
        "mauritius",
        "morocco",
        "mozambique",
        "namibia",
        "niger",
        "nigeria",
        "rwanda",
        "saint_helena_ascension_and_tristan_da_cunha",
        "sao_tome_and_principe",
        "senegal_and_gambia",
        "seychelles",
        "sierra_leone",
        "somalia",
        "south_africa_and_lesotho",
        "south_africa",
        "south_sudan",
        "sudan",
        "swaziland",
        "tanzania",
        "togo",
        "tunisia",
        "uganda",
        "zambia",
        "zimbabwe",
    ]

    available = regions
    available.sort()

    continent = {"name": "africa" + suffix, "url": URL + "africa" + suffix}

    # Create data sources
    _sources = {
        region: {
            "name": region.replace("_", "-") + suffix,
            "url": URL + africa_url + region.replace("_", "-") + suffix,
        }
        for region in regions
    }

    __dict__ = _sources

    def __getattr__(self, name):
        return self.__dict__[name]

    def __call__(self):
        return self.available


class Asia:
    regions = [
        "afghanistan",
        "armenia",
        "azerbaijan",
        "bangladesh",
        "bhutan",
        "cambodia",
        "china",
        "east-timor",
        "gcc_states",
        "india",
        "indonesia",
        "iran",
        "iraq",
        "israel_and_palestine",
        "japan",
        "jordan",
        "kazakhstan",
        "kyrgyzstan",
        "laos",
        "lebanon",
        "malaysia_singapore_brunei",
        "maldives",
        "mongolia",
        "myanmar",
        "nepal",
        "north_korea",
        "pakistan",
        "philippines",
        "south_korea",
        "sri_lanka",
        "syria",
        "taiwan",
        "tajikistan",
        "thailand",
        "turkmenistan",
        "uzbekistan",
        "vietnam",
        "yemen",
    ]

    japan = Japan()

    available = regions
    available.sort()

    continent = {"name": "asia" + suffix, "url": URL + "asia" + suffix}

    # Create data sources
    _sources = {
        region: {
            "name": region.replace("_", "-") + suffix,
            "url": URL + asia_url + region.replace("_", "-") + suffix,
        }
        for region in regions
    }

    __dict__ = _sources

    def __getattr__(self, name):
        return self.__dict__[name]

    def __call__(self):
        return self.available


class Antarctica:
    regions = ["antarctica"]
    available = regions
    available.sort()

    continent = {"name": "antarctica" + suffix, "url": URL + "antarctica" + suffix}

    # Create data sources
    _sources = {
        region: {"name": region + suffix, "url": URL + region + suffix}
        for region in regions
    }

    __dict__ = _sources

    def __getattr__(self, name):
        return self.__dict__[name]

    def __call__(self):
        return self.available


class SubRegions:
    def __init__(self):
        self.regions = [
            "brazil",
            "canada",
            "france",
            "germany",
            "great_britain",
            "italy",
            "japan",
            "netherlands",
            "poland",
            "russia",
            "usa",
        ]
        available = self.regions
        available.sort()

        self.brazil = Brazil()
        self.canada = Canada()
        self.france = France()
        self.germany = Germany()
        self.great_britain = GreatBritain()
        self.italy = Italy()
        self.japan = Japan()
        self.netherlands = Netherlands()
        self.poland = Poland()
        self.russia = Russia()
        self.usa = USA()

        self.available = {name: self.__dict__[name].available for name in self.regions}

    def __getattr__(self, name):
        return self.__dict__[name]

    def __call__(self):
        return self.available



#
# Geometry manipulation
#
# Command size
CMD_BITS = 3

# Commands
CMD_MOVE_TO = 1
CMD_LINE_TO = 2
CMD_SEG_END = 7
CMD_FAKE = 0

# Geometry types
UNKNOWN = 0
POINT = 1
LINESTRING = 2
POLYGON = 3


#
# zig-zag encoder and decoder
#
def zig_zag_encode(n):
    """Return the "zig-zag" encoded unsigned integer corresponding to the signed input integer. This encoding is
    used for MVT geometry deltas."""
    return (n << 1) ^ (n >> 31)


def zig_zag_decode(n):
    """Return the signed integer corresponding to the "zig-zag" encoded unsigned input integer. This encoding is
    used for MVT geometry deltas."""
    return (n >> 1) ^ (-(n & 1))


#
# Options management
#

DEFAULT_ENCODE_OPTIONS = {
    "y_coord_down": False,
    "transformer": None,
    "quantize_bounds": None,
    "extents": 4096,
    "on_invalid_geometry": None,
    "check_winding_order": True,
    "max_geometry_validate_tries": 5,
}

DEFAULT_DECODE_OPTIONS = {"y_coord_down": False, "transformer": None, "geojson": True}


def _get_options(layer_options, default_options, global_default_options, operation_name):
    """Get the entire options dictionary filled using: first, the provided `layer_options`, then the provided
    `default_options` and finally filled using the provided `global_default_options`.

    Args:
        layer_options:
            The options for the current layer.

        default_options:
            The default options of the operation.

        global_default_options:
            The global default options for the operation.

        operation_name:
            The name of the current operation.

    Returns:
        The options to use to operate the layer.
    """
    if default_options is None:
        default_options = global_default_options

    if layer_options is None:
        layer_options = default_options

    result = global_default_options.copy()
    result.update(default_options)
    result.update(layer_options)

    result_keys = set(result.keys())
    expected_keys = set(global_default_options.keys())
    extra_keys = result_keys.difference(expected_keys)
    if extra_keys:
        extra_keys_msg = ", ".join(f"{str(x)!r}" for x in sorted(extra_keys))
        raise ValueError(f"The following options are not allowed for {operation_name} a tile: {extra_keys_msg}.")

    return result


def get_encode_options(layer_options, default_options):
    """Get the entire encoding options dictionary filled using: first, the provided `layer_options`, then the provided
    `default_options` and finally filled using the global default options

    Args:
        layer_options:
            The options for the current layer.

        default_options:
            The default options of the encoding operation.

    Returns:
        The options to use for encoding the layer.
    """
    result = _get_options(
        layer_options=layer_options,
        default_options=default_options,
        global_default_options=DEFAULT_ENCODE_OPTIONS,
        operation_name="encoding",
    )

    # Checks on final values
    extents = result["extents"]
    max_geometry_validate_tries = result["max_geometry_validate_tries"]
    if extents <= 0:
        raise ValueError(f"The extents must be positive. {extents} provided.")
    if max_geometry_validate_tries <= 0:
        raise ValueError(f"The max_geometry_validate_tries must be positive. {max_geometry_validate_tries} provided.")

    return result


def get_decode_options(layer_options, default_options):
    """Get the entire decoding options dictionary filled using: first, the provided `layer_options`, then the provided
    `default_options` and finally filled using the global default options

    Args:
        layer_options:
            The options for the current layer.

        default_options:
            The default options of the decoding operation.

    Returns:
        The options to use for decoding the layer.
    """
    return _get_options(
        layer_options=layer_options,
        default_options=default_options,
        global_default_options=DEFAULT_DECODE_OPTIONS,
        operation_name="decoding",
    )



import requests
import json
import argparse
import sqlite3
import os
from vtiles.utils.vt2geojson.tools import vt_bytes_to_geojson, _is_url
import gzip, zlib
import logging
from re import search
from urllib.request import urlopen

def tile_data_to_geojson(tile_data, x, y, z, output):
    try:
        features = vt_bytes_to_geojson(tile_data, x,y,z)
        with open(output, 'w') as f:
            json.dump(features, f, indent=2)
        print(f"GeoJSON data has been saved to {output}")
    except Exception as e:
        print(f"Error saving pbf to geojson: {e}")


def main():
    parser = argparse.ArgumentParser(description='Convert tile data from BBF file, MBTiles file, or URL to GeoJSON.')
    parser.add_argument('-i', '--input', type=str, required=True, help='Input PBF file, MBTiles file, or URL')
    parser.add_argument('-z', type=int, help='Tile zoom level')
    parser.add_argument('-x', type=int, help='Tile column')
    parser.add_argument('-y', type=int, help='Tile row')
    # parser.add_argument("-l", "--layer", help="include only the specified layer", type=str)
    parser.add_argument('-o', '--output', type=str, required=True, help='Output GeoJSON file')
    parser.add_argument('-flipy', '--flipy', type=int, choices=[0, 1], default=0, help='TMS <--> XYZ tiling scheme (optional): 1 or 0, default is 0')

    args = parser.parse_args()    
    tile_data,z,x,y = None,None,None,None
    url_or_path = args.input
    flipy = args.flipy
    XYZ_REGEX = r"\/(\d+)\/(\d+)\/(\d+)"

    if _is_url(url_or_path):
        matches = search(XYZ_REGEX, args.input)
        if matches is None:
            raise ValueError("Unknown url, no `/z/x/y` pattern.")
        z, x, y = map(int, matches.groups()) # set z, x, y based on /z/x/y values in the URL
        r = urlopen(url_or_path)
        tile_data = r.read()
    else:
        if args.z is None or args.x is None or args.y is None:
            raise ValueError("-z, -x, -y must be specified for MBTiles or PBF file!")
        z = args.z
        x = args.x
        y = args.y        
        if url_or_path.endswith('.mbtiles'):
            try:
                conn = sqlite3.connect(url_or_path)
                cursor = conn.cursor()                   
                cursor.execute('''
                    SELECT tile_data 
                    FROM tiles 
                    WHERE zoom_level=? 
                    AND tile_column=? 
                    AND tile_row=?''',                    
                    (z, x, y))
                row = cursor.fetchone()
                conn.close()
                if row:
                    tile_data = row[0]
                else:
                    logging.error(f"Tile not found in MBTiles file at zoom_level={z}, tile_column={x}, tile_row={y}")
                    return None
            except sqlite3.Error as e:
                logging.error(f"Failed to read MBTiles file {url_or_path}: {e}")
                return None
        
        elif  url_or_path.endswith('.pbf'):
            try:
                with open(url_or_path, 'rb') as f:
                    tile_data = f.read()
            except IOError as e:
                logging.error(f"Failed to read file {url_or_path}: {e}")

    # Decompress tile_data
    if tile_data:
        try:
            if tile_data[:2] == b'\x1f\x8b':  # Check for gzip magic number
                tile_data = gzip.decompress(tile_data)
            elif tile_data[:2] == b'\x78\x9c' or tile_data[:2] == b'\x78\x01' or tile_data[:2] == b'\x78\xda':
                tile_data = zlib.decompress(tile_data) 
        except Exception as e:
            logging.error(f"Failed to decompress gzip data: {e}")
        if flipy: 
            y = (1 << z) - 1 - y          
        tile_data_to_geojson(tile_data, x,y,z,args.output)

if __name__ == '__main__':
    main()



import osmium as o
import sys
import datetime as dt
import osmium.replication.server as rserv
import argparse

class Stats(object):

    def __init__(self):
        self.added = 0
        self.modified = 0
        self.deleted = 0

    def add(self, o):
        if o.deleted:
            self.deleted += 1
        elif o.version == 1:
            self.added += 1
        else:
            self.modified += 1

    def outstats(self, prefix):
        print("%s added: %d" % (prefix, self.added))
        print("%s modified: %d" % (prefix, self.modified))
        print("%s deleted: %d" % (prefix, self.deleted))

class FileStatsHandler(o.SimpleHandler):
    def __init__(self):
        super(FileStatsHandler, self).__init__()
        self.nodes = Stats()
        self.ways = Stats()
        self.rels = Stats()

    def node(self, n):
        self.nodes.add(n)

    def way(self, w):
        self.ways.add(w)

    def relation(self, r):
        self.rels.add(r)

def main():
    parser = argparse.ArgumentParser(description="Compute stats from OSM replication server.")
    parser.add_argument('-url', required=True, help='Replication server URL')
    parser.add_argument('-starttime', required=True, help='Start time in the format YYYY-MM-DDTHH:MM:SSZ')
    parser.add_argument('-maxkB', type=int, required=True, help='Maximum kB to download')

    args = parser.parse_args()

    start = dt.datetime.strptime(args.starttime, "%Y-%m-%dT%H:%M:%SZ")
    if sys.version_info >= (3, 0):
        start = start.replace(tzinfo=dt.timezone.utc)
    maxkb = min(args.maxkB, 10 * 1024)

    repserv = rserv.ReplicationServer(args.url)

    seqid = repserv.timestamp_to_sequence(start)
    print("Initial sequence id:", seqid)

    h = FileStatsHandler()
    seqid = repserv.apply_diffs(h, seqid, maxkb)
    print("Final sequence id:", seqid)

    h.nodes.outstats("Nodes")
    h.ways.outstats("Ways")
    h.rels.outstats("Relations")

if __name__ == '__main__':
    main()



# Reference: 
# https://github.com/mapbox/mbtiles-spec/blob/master/1.3/spec.md
# https://github.com/mapbox/tippecanoe/blob/master/main.cpp#L2033

import sqlite3
import os, sys
import logging
from vtiles.utils.geopreocessing import check_vector, determine_tileformat, get_zoom_levels,get_bounds_center
logging.basicConfig(level=logging.INFO)


def fix_vectormetadata(input_mbtiles, compression_type):
    conn = sqlite3.connect(input_mbtiles)       
    cursor = conn.cursor()
    cursor.execute('CREATE TABLE IF NOT EXISTS metadata (name TEXT, value TEXT);')
    cursor.execute('create unique index IF NOT EXISTS name on metadata (name);')    

    # Update name and description
    name = os.path.basename(input_mbtiles)
    cursor.execute("INSERT OR IGNORE INTO metadata (name, value) VALUES (?, ?)", ('name', name))
    desc = 'Update metadata by vtiles.mbtiles.fixmeta'
    cursor.execute("INSERT OR IGNORE INTO metadata (name, value) VALUES (?, ?)", ('description', desc))

    # Update format to pbf 
    cursor.execute("INSERT OR REPLACE INTO metadata (name, value) VALUES (?, ?)", ('format', 'pbf'))
   
    # Update compression
    cursor.execute("INSERT OR REPLACE INTO metadata (name, value) VALUES (?, ?)", ('compression', compression_type))
    
    # Update min zoom, max zoom
    min_zoom, max_zoom = get_zoom_levels(input_mbtiles)
    cursor.execute("INSERT OR REPLACE INTO metadata (name, value) VALUES (?, ?)", ('minzoom', min_zoom))
    cursor.execute("INSERT OR REPLACE INTO metadata (name, value) VALUES (?, ?)", ('maxzoom', max_zoom))

    # Update bounds and center
    bounds, center = get_bounds_center(input_mbtiles)
    if bounds:
        cursor.execute("INSERT OR REPLACE INTO metadata (name, value) VALUES (?, ?)", ('bounds', bounds))
    if center:
        cursor.execute("INSERT OR REPLACE INTO metadata (name, value) VALUES (?, ?)", ('center', center))
    conn.commit()
    conn.close() 
    print(f'Fix metadata for {name} done!')

def fix_rastermetadata(input_mbtiles, format):
    conn = sqlite3.connect(input_mbtiles)       
    cursor = conn.cursor()
    cursor.execute('CREATE TABLE IF NOT EXISTS metadata (name TEXT, value TEXT);')
    cursor.execute('create unique index IF NOT EXISTS name on metadata (name);')    

    # Update name and description
    name = os.path.basename(input_mbtiles)
    cursor.execute("INSERT OR IGNORE INTO metadata (name, value) VALUES (?, ?)", ('name', name))
    desc = 'Update metadata by vtiles.mbtiles.fixmeta'
    cursor.execute("INSERT OR IGNORE INTO metadata (name, value) VALUES (?, ?)", ('description', desc))
   
    # Update format to raster
    cursor.execute("INSERT OR REPLACE INTO metadata (name, value) VALUES (?, ?)", ('format',format))

  
    # Update min zoom, max zoom
    min_zoom, max_zoom = get_zoom_levels(input_mbtiles)
    cursor.execute("INSERT OR REPLACE INTO metadata (name, value) VALUES (?, ?)", ('minzoom', min_zoom))
    cursor.execute("INSERT OR REPLACE INTO metadata (name, value) VALUES (?, ?)", ('maxzoom', max_zoom))

    # Update bounds and center
    bounds, center = get_bounds_center(input_mbtiles)
    if bounds:
        cursor.execute("INSERT OR REPLACE INTO metadata (name, value) VALUES (?, ?)", ('bounds', bounds))
    if center:
        cursor.execute("INSERT OR REPLACE INTO metadata (name, value) VALUES (?, ?)", ('center', center))
    conn.commit()
    conn.close() 
    print(f'Fix metadata for {name} done!')

def main():
    if len(sys.argv) != 2:
      print("Please provide the MBTiles input filename.")
      return
    input_mbtiles = sys.argv[1]
    if (os.path.exists(input_mbtiles)):
        is_vector, compression_type = check_vector(input_mbtiles) 
        tile_format = determine_tileformat(input_mbtiles) 
        if is_vector:
            fix_vectormetadata(input_mbtiles, compression_type)   
        else:
            fix_rastermetadata(input_mbtiles, tile_format)        
    else: 
        print ('MBTiles file does not exist!. Please recheck and input a correct file path.')
        return

if __name__ == "__main__":
    main()



# python setup.py sdist bdist_wheel
# twine upload dist/*

from setuptools import setup, find_packages

requirements = [
    'tqdm~=4.66.2',
    'boto3~=1.34.72',
    'requests~=2.31.0',
    'click~=8.1.7',
    'fiona~=1.10.0',
    'shapely~=2.0.1',
    'osmium~=3.7.0',
    'protobuf~=5.26.1',
    'ujson'
],

setup(
    name='vtiles',
    version='1.0.4',
    author = 'Thang Quach',
    author_email= 'quachdongthang@gmail.com',
    url='https://github.com/thangqd/vtiles',
    description='Vector Tiles Utilities',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    requires_python=">=3.0",
    packages=find_packages(),
    entry_points={
        'console_scripts': [
            'osmdownload = vtiles.osm.osmdownload:main',
            'osmreplication = vtiles.osm.osmreplication:main',
            'osminfo = vtiles.osm.osminfo:main',
            'osm2geojson = vtiles.osm.osm2geojson:main',
            'osmpub = vtiles.osm.osmpub:main',
            
            'mbtilesinfo = vtiles.mbtiles.mbtilesinfo:main',
            'mbtilesinspect = vtiles.mbtiles.mbtilesinspect:main',
            'mbtilesdelduplicate = vtiles.mbtiles.mbtilesdelduplicate:main',           

            'mbtiles2folder = vtiles.mbtiles.mbtiles2folder:main',
            'folder2mbtiles = vtiles.mbtiles.folder2mbtiles:main',
            'url2folder = vtiles.mbtiles.url2folder:main',   
            'mbtiles2geojson = vtiles.mbtiles.mbtiles2geojson:main',
            'geojson2mbtiles = vtiles.mbtiles.geojson2mbtiles:main',                   
            'mbtiles2s3 = vtiles.mbtiles.mbtiles2s3:main',        
            'folder2s3 = vtiles.mbtiles.folder2s3:main',         
            'mbtiles2pbf = vtiles.mbtiles.mbtiles2pbf:main',

            'mbtilessplit = vtiles.mbtiles.mbtilessplit:main',
            'mbtilesmerge = vtiles.mbtiles.mbtilesmerge:main',
            'mbtilesdecompress = vtiles.mbtiles.mbtilesdecompress:main',
            'mbtilescompress = vtiles.mbtiles.mbtilescompress:main',
            'mbtilesfixmeta = vtiles.mbtiles.mbtilesfixmeta:main',
           
            'pbfinfo = vtiles.mbtiles.pbfinfo:main',
            'pbf2geojson = vtiles.mbtiles.pbf2geojson:main',      
            'flipy = vtiles.mbtiles.flipy:main',   
            'mbtiles2pmtiles = vtiles.mbtiles.mbtiles2pmtiles:main',         
            
            'servefolder= vtiles.server.servefolder:main',
            'servepostgis= vtiles.server.servepostgis:main',          
            'servembtiles=vtiles.server.servembtiles:main',    
            'serverastermbtiles=vtiles.server.serverastermbtiles:main',
            'servevectormbtiles=vtiles.server.servevectormbtiles:main',       
            'servepmtiles=vtiles.server.servepmtiles:main',                   
            
           
            'pmtilesinfo = vtiles.utils.pmtilesinfo:main',
            'pmtiles2folder = vtiles.utils.pmtiles2folder:main',
            'pmtiles2mbtiles = vtiles.utils.pmtiles2mbtiles:main',           
            'vtpk2folder=vtiles.utils.vtpk2folder:main',
            'centerline=vtiles.utils.centerline:main'
        ],
    },    

    # scripts=["bin/utils.py"], # utils.py is just a demo,
    install_requires=requirements,    
    classifiers=[
        'Programming Language :: Python :: 3',
        'Environment :: Console',
        'Topic :: Scientific/Engineering :: GIS',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
)



import osmium as o
import argparse

class PubHandler(o.SimpleHandler):
    def __init__(self):
        super(PubHandler, self).__init__()
        self.pubs = []

    def node(self, n):
        if 'amenity' in n.tags and n.tags['amenity'] == 'pub' and 'name' in n.tags:
            self.pubs.append(n.tags['name'])

    def way(self, w):
        if 'amenity' in w.tags and w.tags['amenity'] == 'pub' and 'name' in w.tags:
            self.pubs.append(w.tags['name'])

    def relation(self, r):
        if 'amenity' in r.tags and r.tags['amenity'] == 'pub' and 'name' in r.tags:
            self.pubs.append(r.tags['name'])

def main():
    parser = argparse.ArgumentParser(description="Search for pubs in an OSM file and list their names.")
    parser.add_argument('-i', '--input', required=True, help='Path to the input OSM file')

    args = parser.parse_args()

    handler = PubHandler()
    handler.apply_file(args.input)

    for pub_name in handler.pubs:
        print(pub_name)

if __name__ == '__main__':
    main()
